from typing import List, Literal

from ....utils.class_registry import register_class

from ..character_base import SkillTalent
from ...consts import DieColor
from ...struct import Cost
from .yoimiya_3_8 import Yoimiya_3_8 as Y_3_8
from .yoimiya_3_8 import RyuukinSaxifrage as RS_3_8
from .yoimiya_3_8 import FireworkFlareUp, NiwabiFireDance


class RyuukinSaxifrage(RS_3_8):
    cost: Cost = Cost(
        elemental_dice_color=DieColor.PYRO, elemental_dice_number=3, charge=2
    )


class NaganoharaMeteorSwarm_3_3(SkillTalent):
    name: Literal["Naganohara Meteor Swarm"]
    version: Literal["3.3"] = "3.3"
    character_name: Literal["Yoimiya"] = "Yoimiya"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.PYRO,
        elemental_dice_number=2,
    )
    skill: Literal["Niwabi Fire-Dance"] = "Niwabi Fire-Dance"
    status_max_usage: int = 2


class Yoimiya_3_3(Y_3_8):
    version: Literal["3.3"]
    max_charge: int = 2
    skills: List[FireworkFlareUp | NiwabiFireDance | RyuukinSaxifrage] = []

    def _init_skills(self) -> None:
        self.skills = [FireworkFlareUp(), NiwabiFireDance(), RyuukinSaxifrage()]


register_class(Yoimiya_3_3 | NaganoharaMeteorSwarm_3_3)



from typing import Dict, List, Literal
from lpsim.server.action import ChangeObjectUsageAction
from lpsim.server.match import Match
from lpsim.server.object_base import EventCardBase
from lpsim.server.struct import Cost, ObjectPosition
from lpsim.utils.class_registry import register_class
from lpsim.utils.desc_registry import DescDictType


class ControlledDirectionalBlast_4_5(EventCardBase):
    name: Literal["Controlled Directional Blast"] = "Controlled Directional Blast"
    version: Literal["4.5"] = "4.5"
    cost: Cost = Cost(same_dice_number=1)

    def is_valid(self, match: Match) -> bool:
        """
        can use if opponent has 4 or more support and summon and any summon exist
        """
        return (
            len(self.query(match, "opponent support and opponent summon")) >= 4
            and len(self.query(match, "both summon")) > 0
        )

    def get_targets(self, match: Match) -> List[ObjectPosition]:
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Match
    ) -> List[ChangeObjectUsageAction]:
        assert target is None
        targets = self.query(match, "both summon")
        ret: List[ChangeObjectUsageAction] = []
        for t in targets:
            ret.append(
                ChangeObjectUsageAction(object_position=t.position, change_usage=-1)
            )
        return ret


desc: Dict[str, DescDictType] = {
    "CARD/Controlled Directional Blast": {
        "names": {
            "en-US": "Controlled Directional Blast",
            "zh-CN": "可控性去危害化式定向爆破",
        },
        "descs": {
            "4.5": {
                "en-US": "Can only be played when there is a total of at least 4 cards in your opponent's Support Zone and Summons Zone: All Summons on both sides lose 1 Usage.",  # noqa: E501
                "zh-CN": "对方支援区和召唤物区的卡牌数量总和至少为4时，才能打出：双方所有召唤物的可用次数-1。",
            }
        },
        "image_path": "cardface/Event_Event_Kexueyuan.png",  # noqa: E501
        "id": 332030,
    },
}


register_class(ControlledDirectionalBlast_4_5, desc)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...event import RoundEndEventArguments

from ...summon.base import AttackerSummonBase

from ...action import (
    ActionTypes,
    Actions,
    ChangeObjectUsageAction,
    CreateObjectAction,
    MakeDamageAction,
)
from ...struct import Cost, ObjectPosition

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Summons


class TenguJuuraiAmbush_3_5(AttackerSummonBase):
    name: Literal["Tengu Juurai: Ambush"] = "Tengu Juurai: Ambush"
    version: Literal["3.5"] = "3.5"
    usage: int = 1
    max_usage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.ELECTRO
    damage: int = 1

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | CreateObjectAction | ChangeObjectUsageAction]:
        ret: List[MakeDamageAction | CreateObjectAction | ChangeObjectUsageAction] = []
        ret += super().event_handler_ROUND_END(event, match)
        damage_action = ret[0]
        assert damage_action.type == ActionTypes.MAKE_DAMAGE
        active_idx = match.player_tables[self.position.player_idx].active_character_idx
        ret.append(
            CreateObjectAction(
                object_name="Crowfeather Cover",
                object_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    character_idx=active_idx,
                    area=ObjectPositionType.CHARACTER_STATUS,
                    id=-1,
                ),
                object_arguments={},
            )
        )
        return ret


class TenguJuuraiStormcluster_3_5(TenguJuuraiAmbush_3_5):
    name: Literal["Tengu Juurai: Stormcluster"] = "Tengu Juurai: Stormcluster"
    usage: int = 2
    max_usage: int = 2
    damage: int = 2


# Skills


class TenguStormcall(ElementalSkillBase):
    name: Literal["Tengu Stormcall"] = "Tengu Stormcall"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [self.create_summon("Tengu Juurai: Ambush")]


class SubjugationKoukouSendou(ElementalBurstBase):
    name: Literal["Subjugation: Koukou Sendou"] = "Subjugation: Koukou Sendou"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=4, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_summon("Tengu Juurai: Stormcluster")
        ]


# Talents


class SinOfPride_3_5(SkillTalent):
    name: Literal["Sin of Pride"]
    version: Literal["3.5"] = "3.5"
    character_name: Literal["Kujou Sara"] = "Kujou Sara"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=4, charge=2
    )
    skill: Literal["Subjugation: Koukou Sendou"] = "Subjugation: Koukou Sendou"


class SinOfPride_4_2(SkillTalent):
    name: Literal["Sin of Pride"]
    version: Literal["4.2"] = "4.2"
    character_name: Literal["Kujou Sara"] = "Kujou Sara"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO,
        elemental_dice_number=3,
    )
    skill: Literal["Tengu Stormcall"] = "Tengu Stormcall"


# character base


class KujouSara_3_5(CharacterBase):
    name: Literal["Kujou Sara"]
    version: Literal["3.5"] = "3.5"
    element: ElementType = ElementType.ELECTRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | TenguStormcall | SubjugationKoukouSendou
    ] = []
    faction: List[FactionType] = [FactionType.INAZUMA]
    weapon_type: WeaponType = WeaponType.BOW

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Tengu Bowmanship",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            TenguStormcall(),
            SubjugationKoukouSendou(),
        ]


register_class(
    KujouSara_3_5
    | SinOfPride_3_5
    | SinOfPride_4_2
    | TenguJuuraiAmbush_3_5
    | TenguJuuraiStormcluster_3_5
)



from typing import Dict, Literal

from .....utils.class_registry import register_class

from .....utils.desc_registry import DescDictType
from ....character.pyro.fatui_pyro_agent_3_3 import FatuiPyroAgent_3_3


class FatuiPyroAgent_4_3(FatuiPyroAgent_3_3):
    name: Literal["Fatui Pyro Agent"]
    version: Literal["4.3"] = "4.3"
    max_hp: int = 9


desc: Dict[str, DescDictType] = {
    "CHARACTER/Fatui Pyro Agent": {
        "descs": {"4.3": {"zh-CN": "", "en-US": ""}},
    },
    "SKILL_Fatui Pyro Agent_ELEMENTAL_BURST/Blade Ablaze": {
        "descs": {
            "4.3": {
                "zh-CN": "$SKILL_Fatui Pyro Agent_ELEMENTAL_BURST/Blade Ablaze|descs|3.3|zh-CN",  # noqa: E501
                "en-US": "$SKILL_Fatui Pyro Agent_ELEMENTAL_BURST/Blade Ablaze|descs|3.3|en-US",  # noqa: E501
            }
        }
    },
    "SKILL_Fatui Pyro Agent_ELEMENTAL_SKILL/Prowl": {
        "descs": {
            "4.3": {
                "zh-CN": "$SKILL_Fatui Pyro Agent_ELEMENTAL_SKILL/Prowl|descs|3.3|zh-CN",  # noqa: E501
                "en-US": "$SKILL_Fatui Pyro Agent_ELEMENTAL_SKILL/Prowl|descs|3.3|en-US",  # noqa: E501
            }
        }
    },
    "SKILL_Fatui Pyro Agent_NORMAL_ATTACK/Thrust": {
        "descs": {
            "4.3": {
                "zh-CN": "$SKILL_Fatui Pyro Agent_NORMAL_ATTACK/Thrust|descs|3.3|zh-CN",  # noqa: E501
                "en-US": "$SKILL_Fatui Pyro Agent_NORMAL_ATTACK/Thrust|descs|3.3|en-US",  # noqa: E501
            }
        }
    },
    "SKILL_Fatui Pyro Agent_PASSIVE/Stealth Master": {
        "descs": {
            "4.3": {
                "zh-CN": "$SKILL_Fatui Pyro Agent_PASSIVE/Stealth Master|descs|3.3|zh-CN",  # noqa: E501
                "en-US": "$SKILL_Fatui Pyro Agent_PASSIVE/Stealth Master|descs|3.3|en-US",  # noqa: E501
            }
        }
    },
}


register_class(FatuiPyroAgent_4_3, desc)



from typing import List, Literal

from ....utils.class_registry import register_class
from ...consts import (
    ElementType,
    FactionType,
    WeaponType,
    DamageElementalType,
)
from ..character_base import (
    PhysicalNormalAttackBase,
    ElementalSkillBase,
    ElementalBurstBase,
    CharacterBase,
)


class PhysicalMob_1_0(CharacterBase):
    """
    Physical mobs. They cannot carry weapons in default. Their normal
    attacks are 2 physical DMG, elemental skills are 3 physical DMG,
    elemental bursts are 2 charges, 3 cost, and 5 physical DMG.
    Their elemeny type is only used to decide dice color when using skills.
    """

    name: Literal["PhysicalMob"]
    version: Literal["1.0"] = "1.0"
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | ElementalSkillBase | ElementalBurstBase
    ] = []
    faction: List[FactionType] = []
    weapon_type: WeaponType = WeaponType.OTHER

    def __init__(self, **kwargs):
        super().__init__(**kwargs)  # type: ignore

    def _init_skills(self):
        element_name = self.element.value.lower()
        element_name = element_name[0].upper() + element_name[1:]
        normal_attack = PhysicalNormalAttackBase(
            name="Physical Normal Attack",
            damage_type=DamageElementalType.PHYSICAL,
            cost=PhysicalNormalAttackBase.get_cost(self.element),
        )
        elemental_skill = ElementalSkillBase(
            name="Physical Skill",
            damage_type=DamageElementalType.PHYSICAL,
            cost=ElementalSkillBase.get_cost(self.element),
        )
        elemental_burst = ElementalBurstBase(
            name="Physical Burst",
            damage_type=DamageElementalType.PHYSICAL,
            cost=ElementalBurstBase.get_cost(self.element, 3, 2),
            damage=5,
        )
        self.skills = [normal_attack, elemental_skill, elemental_burst]


register_class(PhysicalMob_1_0)



from typing import Any, Dict, List, Literal
from lpsim.server.action import (
    Actions,
    CreateDiceAction,
    DrawCardAction,
    RemoveObjectAction,
)
from lpsim.server.card.support.base import UsageWithRoundRestrictionSupportBase
from lpsim.server.card.support.items import ItemBase
from lpsim.server.consts import DieColor, IconType, PlayerActionLabels
from lpsim.server.event import ActionEndEventArguments, RoundPrepareEventArguments
from lpsim.server.match import Match
from lpsim.server.struct import Cost
from lpsim.utils.class_registry import register_class
from lpsim.utils.desc_registry import DescDictType


class Lumenstonedjuvant_4_5(ItemBase, UsageWithRoundRestrictionSupportBase):
    name: Literal["Lumenstone Adjuvant"] = "Lumenstone Adjuvant"
    version: Literal["4.5"] = "4.5"
    cost: Cost = Cost(same_dice_number=2)
    max_usage_one_round: int = 1
    usage: int = 3
    counter: int = 0
    icon_type: IconType = IconType.COUNTER

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        self.counter = 0
        return super().event_handler_ROUND_PREPARE(event, match)

    def event_handler_ACTION_END(
        self, event: ActionEndEventArguments, match: Match
    ) -> List[CreateDiceAction | DrawCardAction | RemoveObjectAction]:
        """
        When used 3 cards, create 1 omni die and draw one card.
        """
        if not self.has_usage():
            return []
        if self.position.not_satisfy(
            "source area=support and both player=same id=diff", event.action.position
        ):
            return []
        if event.action.action_label & PlayerActionLabels.CARD != 0:
            # is card action
            self.counter += 1
        if self.counter < 3:
            return []
        self.use()
        return [
            CreateDiceAction(
                player_idx=self.position.player_idx,
                number=1,
                color=DieColor.OMNI,
            ),
            DrawCardAction(
                player_idx=self.position.player_idx,
                number=1,
                draw_if_filtered_not_enough=True,
            ),
        ] + self.check_should_remove()


desc: Dict[str, DescDictType] = {
    "SUPPORT/Lumenstone Adjuvant": {
        "names": {"en-US": "Lumenstone Adjuvant", "zh-CN": "流明石触媒"},
        "descs": {
            "4.5": {
                "en-US": "After you play an Action Card: If, while this card is on the field, you have already played 3 Action Cards this Round, draw 1 card and create 1 Omni Die. (Once per Round)\nUsage(s): 3",  # noqa: E501
                "zh-CN": "我方打出行动牌后：如果此牌在场期间本回合中我方已打出3张行动牌，则抓1张牌并生成1个万能元素骰。（每回合1次）\n可用次数：3",  # noqa: E501
            }
        },
        "image_path": "cardface/Assist_Prop_Liumingshi.png",  # noqa: E501
        "id": 323007,
    },
}


register_class(Lumenstonedjuvant_4_5, desc)



from typing import Any, List, Literal

from .....utils.class_registry import register_class

from ....consts import DamageElementalType, DamageType, ObjectPositionType

from ....action import Actions, MakeDamageAction

from ....event import MakeDamageEventArguments

from ....modifiable_values import DamageIncreaseValue, DamageValue
from .base import ArtifactBase
from ....struct import Cost


class CrownOfWatatsumi_4_1(ArtifactBase):
    name: Literal["Crown of Watatsumi"]
    version: Literal["4.1"] = "4.1"
    cost: Cost = Cost(same_dice_number=1)
    usage: int = 0
    max_usage: int = 2
    counter: int = 0

    def equip(self, match: Any) -> List[Actions]:
        self.usage = 0
        self.counter = 0
        return []

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, None
        ):
            # not self use damage skill
            return value
        # increase damage
        assert mode == "REAL"
        value.damage += self.usage
        self.usage = 0
        return value

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[Actions]:
        """
        If our player receives heal, add counter.
        """
        if self.position.area != ObjectPositionType.CHARACTER:
            # not equipped
            return []
        for damage in event.damages:
            if (
                damage.final_damage.target_position.player_idx
                != self.position.player_idx
                or damage.final_damage.damage_type != DamageType.HEAL
            ):
                # target not self player, or not heal
                continue
            # increase counter by heal
            self.counter += damage.hp_after - damage.hp_before
            # self.counter += - damage.final_damage.damage
            self.usage += self.counter // 3
            self.counter = self.counter % 3
            if self.usage >= self.max_usage:
                # usage full
                self.counter = 0
                self.usage = self.max_usage
        return []


class OceanHuedClam_4_2(CrownOfWatatsumi_4_1):
    name: Literal["Ocean-Hued Clam"]
    version: Literal["4.2"] = "4.2"
    cost: Cost = Cost(any_dice_number=3)

    def equip(self, match: Any) -> List[Actions]:
        super().equip(match)
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        # heal self
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        target_position=character.position,
                        damage_type=DamageType.HEAL,
                        damage=-3,
                        damage_elemental_type=DamageElementalType.HEAL,
                        cost=Cost(),
                    )
                ]
            )
        ]


register_class(CrownOfWatatsumi_4_1 | OceanHuedClam_4_2)



from typing import Dict, List, Literal

from ....character.character_base import SkillBase
from ....event import SkillEndEventArguments
from ....status.character_status.base import RoundCharacterStatus
from ....match import Match
from ....card.equipment.weapon.base import WeaponBase
from ....action import CreateDiceAction, CreateObjectAction, RemoveObjectAction
from ....consts import (
    ELEMENT_TO_DIE_COLOR,
    IconType,
    ObjectPositionType,
    SkillType,
    WeaponType,
)
from ....struct import Cost
from .....utils.class_registry import register_class
from .....utils.desc_registry import DescDictType


class SapwoodBladeStatus_4_4(RoundCharacterStatus):
    name: Literal["Sapwood Blade"]
    version: Literal["4.4"] = "4.4"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.BUFF] = IconType.BUFF

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Match
    ) -> List[CreateDiceAction | RemoveObjectAction]:
        """
        When self use normal attack, create 2 elemental dice.
        """
        position = event.action.position
        if not self.position.check_position_valid(
            position,
            match,
            player_idx_same=True,
            character_idx_same=True,
        ):
            # not self use skill
            return []
        skill: SkillBase = match.get_object(position)  # type: ignore
        if skill.skill_type != SkillType.NORMAL_ATTACK:
            # not using normal attack
            return []
        character = match.player_tables[position.player_idx].characters[
            position.character_idx
        ]
        assert self.usage > 0
        self.usage -= 1
        return [
            CreateDiceAction(
                player_idx=position.player_idx,
                color=ELEMENT_TO_DIE_COLOR[character.element],
                number=2,
            )
        ] + self.check_should_remove()


class SapwoodBlade_4_4(WeaponBase):
    name: Literal["Sapwood Blade"]
    version: Literal["4.4"] = "4.4"
    weapon_type: WeaponType = WeaponType.SWORD
    cost: Cost = Cost(any_dice_number=3)

    def equip(self, match: Match) -> List[CreateObjectAction]:
        """
        attach status
        """
        return [
            CreateObjectAction(
                object_position=self.position.set_area(
                    ObjectPositionType.CHARACTER_STATUS
                ),
                object_name=self.name,
                object_arguments={},
            )
        ]


desc: Dict[str, DescDictType] = {
    "CHARACTER_STATUS/Sapwood Blade": {
        "names": {"en-US": "Sapwood Blade", "zh-CN": "原木刀"},
        "descs": {
            "4.4": {
                "en-US": "When this character next uses a Normal Attack during this Round: Create 2 Elemental Dice of this character's Elemental Type.",  # noqa: E501
                "zh-CN": "本回合中，下次使用「普通攻击」后：生成2个此角色类型的元素骰。",  # noqa: E501
            }
        },
    },
    "WEAPON/Sapwood Blade": {
        "names": {"en-US": "Sapwood Blade", "zh-CN": "原木刀"},
        "descs": {
            "4.4": {
                "en-US": "The character deals +1 DMG.\nWhen entering play: When the character to which this is attached next uses a Normal Attack during this Round: Create 2 Elemental Dice of this character's Elemental Type.\n(Only Sword Characters can equip this. A character can equip a maximum of 1 Weapon)",  # noqa: E501
                "zh-CN": "角色造成的伤害+1。\n入场时：所附属角色在本回合中，下次使用「普通攻击」后：生成2个此角色类型的元素骰。\n（「单手剑」角色才能装备。角色最多装备1件「武器」）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Weapon_Yuanmu.png",  # noqa: E501
        "id": 311507,
    },
}


register_class(SapwoodBlade_4_4 | SapwoodBladeStatus_4_4, desc)



from lpsim.server.status.character_status.geo_characters import RagingOniKing_3_6
from lpsim.agents.interaction_agent import InteractionAgent
from lpsim.server.match import Match, MatchState
from lpsim.server.deck import Deck
from tests.utils_for_test import (
    check_hp,
    check_usage,
    get_pidx_cidx,
    get_random_state,
    get_test_id_from_command,
    make_respond,
    set_16_omni,
)


def test_arataki_itto():
    cmd_records = [
        [
            "sw_card",
            "choose 2",
            "end",
            "sw_char 0 0",
            "sw_char 1 0",
            "sw_char 2 0",
            "sw_char 1 0",
            "sw_char 2 0",
            "TEST 1 10 9 78 10 10 90",
            "sw_char 1 0",
            "skill 1 0 1 2",
            "sw_char 2 0",
            "end",
            "end",
            "end",
            "sw_char 1 0",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "skill 1 0 1 2",
            "choose 2",
            "TEST 1 0 9 40 7 8 90",
            "TEST 2 p1c0 status 2 usage 2 2",
            "end",
        ],
        [
            "sw_card",
            "choose 0",
            "skill 0 0 1 2",
            "skill 1 0 1 2",
            "TEST 3 skill 0 cost 3",
            "skill 0 0 1 2",
            "TEST 2 p1c0 status 0 usage ",
            "TEST 1 10 10 84 10 10 90",
            "skill 1 0 1 2",
            "skill 1 0 1 2",
            "TEST 2 p1c0 status 1 usage 2",
            "end",
            "TEST 3 skill 0 cost 2",
            "skill 1 0 1 2",
            "TEST 3 skill 0 cost 3",
            "sw_char 1 0",
            "sw_char 2 0",
            "sw_char 0 0",
            "TEST 3 skill 0 cost 2",
            "skill 0 0 1",
            "sw_char 1 0",
            "TEST 1 10 9 78 10 8 90",
            "TEST 2 p1c0 status 1 usage 3",
            "sw_char 0 0",
            "end",
            "skill 2 0 1 2",
            "sw_char 1 0",
            "sw_char 2 0",
            "sw_char 0 0",
            "TEST 3 skill 0 cost 2",
            "TEST 2 p1c0 status 2 usage 3 2",
            "skill 0 0 1",
            "TEST 1 10 9 67 10 8 90",
            "TEST 2 p1c0 status 2 usage 3 2",
            "skill 0 0 1",
            "TEST 2 p1c0 status 2 usage 2 2",
            "skill 0 0 1",
            "skill 2 0 1 2",
            "TEST 4 burst status renew usage",
            "end",
            "skill 1 0 1 2",
            "TEST 2 p1c0 status 2 usage 2 1",
            "skill 0 0 1 2",
            "TEST 2 p1c0 status 2 usage 3 1",
            "TEST 5 card cost 3",
            "card 0 0 0 1 2",
            "TEST 1 10 9 41 10 8 90",
            "end",
            "skill 2 0 1 2",
            "skill 0 0 1 2",
            "TEST 2 p1c0 status 2 usage 3 2",
            "skill 0 0 1",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Arataki Itto
        character:PhysicalMob
        character:Mona
        Arataki Ichiban*30
        """
    )
    deck.characters[2].hp = 90
    deck.characters[2].max_hp = 90
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                cmd = cmd.strip().split()
                pc = cmd[2]
                status = int(cmd[4])
                usage = [int(x) for x in cmd[6:]]
                pidx = int(pc[1])
                cidx = int(pc[3])
                character = match.player_tables[pidx].characters[cidx]
                assert len(character.status) == status
                for s, u in zip(character.status, usage):
                    assert s.usage == u
            elif test_id == 3:
                cmd = cmd.strip().split()
                sidx = int(cmd[3])
                cost = int(cmd[5])
                for req in match.requests:
                    if req.name == "UseSkillRequest":
                        if req.skill_idx == sidx:
                            assert req.cost.total_dice_cost == cost
                            assert req.cost.elemental_dice_number == 1
            elif test_id == 4:
                character = match.player_tables[1].characters[0]
                for status in character.status:
                    if status.name == "Raging Oni King":
                        assert isinstance(status, RagingOniKing_3_6)
                        assert status.status_increase_usage == 1
            elif test_id == 5:
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        assert req.cost.total_dice_cost == 3
                        assert req.cost.elemental_dice_number == 1
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_arataki_itto_2():
    cmd_records = [
        [
            "sw_card",
            "choose 2",
            "end",
            "end",
            "end",
            "skill 1 0 1 2",
            "skill 1 0 1 2",
        ],
        [
            "sw_card 0 1 2 3",
            "choose 1",
            "skill 1 0 1 2",
            "skill 0 0 1 2",
            "card 0 0 0 1 2",
            "TEST 1 10 10 83 10 10 90",
            "end",
            "card 0 1 0",
            "skill 1 0 1 2",
            "skill 1 0 1 2",
            "skill 0 0 1 2",
            "skill 0 0 1",
            "TEST 1 10 10 74 10 10 90",
            "end",
            "skill 0 0 1 2",
            "TEST 1 10 10 70 10 10 90",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "TEST 1 10 10 66 10 10 90",
            "end",
            "skill 1 0 1 2",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:PhysicalMob
        character:Arataki Itto
        character:Mona
        Arataki Ichiban*15
        Lotus Flower Crisp*15
        """
    )
    deck.characters[2].hp = 90
    deck.characters[2].max_hp = 90
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_arataki_itto_3():
    cmd_records = [
        [
            "sw_card",
            "choose 1",
            "skill 0 15 14 13",
            "skill 1 12 11 10",
            "card 1 0 9",
            "skill 1 8 7 6",
            "end",
            "TEST 1 p1c1 usage 1 2",
            "TEST 1 p0c1 usage 3 1",
            "end",
        ],
        [
            "sw_card",
            "choose 1",
            "skill 0 15 14 13",
            "sw_char 2 12",
            "sw_char 1 11",
            "card 1 0 10",
            "TEST 1 p0c1 usage 2 1 2",
            "skill 1 9 8 7",
            "TEST 1 p0c1 usage 3 1 2",
            "TEST 1 p1c1 usage 1 2 1",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.1
        character:Beidou
        character:Arataki Itto
        character:Klee
        Mushroom Pizza*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cmd = cmd.strip().split(" ")
                pidx, cidx = get_pidx_cidx(cmd)
                check_usage(match.player_tables[pidx].characters[cidx].status, cmd[4:])
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    # test_arataki_itto()
    # test_arataki_itto_2()
    test_arataki_itto_3()



import json
import numpy as np
from typing import Callable, Dict, List, Any, Tuple
from lpsim.agents.interaction_agent import InteractionAgent, InteractionAgent_V1_0
from lpsim.server.match import Match, MatchConfig, MatchState
from lpsim.server.deck import Deck
from lpsim.server.event_handler import OmnipotentGuideEventHandler_3_3
from lpsim.agents.agent_base import AgentBase
from lpsim.server.struct import ObjectPosition
from lpsim.utils import BaseModel
from tests.default_random_state import get_default_random_state


def set_16_omni(match: Match):
    match.config.initial_dice_number = 16
    match.event_handlers.append(OmnipotentGuideEventHandler_3_3())


def make_respond(agent: AgentBase, match: Match, assertion: bool = True):
    """
    Make response once, and step the match until new request or
    game end, regardless of the target of requests.
    If not requesting this agent and needs assetion, raise AssertionError.
    Otherwise, do nothing.
    """
    resp = agent.generate_response(match)
    if assertion:
        assert resp is not None
    if resp is None:
        return
    match.respond(resp)
    match.step()


def check_hp(match: Match, hp: List[List[int]]):
    """
    input hps for each player and check
    """
    hps = [[x.hp for x in t.characters] for t in match.player_tables]
    assert hps == hp


def check_name(name: str, lists: List[Any], exist: bool = True):
    """
    check the existence of name in lists
    """
    real_exist = name in [x.name for x in lists]
    assert real_exist == exist


def remove_ids(model: BaseModel) -> BaseModel:
    """
    remove ids of all objects in-place.
    """
    for key in model.__fields__.keys():
        value = getattr(model, key)
        if isinstance(value, ObjectPosition):
            model.__setattr__(key, value.set_id(0))
        elif isinstance(value, BaseModel):
            remove_ids(value)
        elif isinstance(value, list) or isinstance(value, tuple):
            assert isinstance(value, list)
            for num, v in enumerate(value):
                if isinstance(v, ObjectPosition):
                    value[num] = v.set_id(0)
                elif isinstance(v, BaseModel):
                    remove_ids(v)
    if "id" in model.__fields__.keys():
        model.id = 0
    return model


def get_random_state(offset: int = 0):
    """
    get random state tuple for test. it will use a fixed random state and
    random offset times as the result random state.
    """
    state: List = get_default_random_state()
    if offset > 0:
        random_state = np.random.RandomState()
        state[1] = np.array(state[1], dtype="uint32")
        random_state.set_state(state)  # type: ignore
        for _ in range(offset):
            random_state.random()
        state = list(random_state.get_state(legacy=True))  # type: ignore
        state[1] = state[1].tolist()
    return state


def get_test_id_from_command(agent: InteractionAgent | InteractionAgent_V1_0):
    """
    from commands to get test number. return test number and remove this
    test message from commands if success. otherwise return 0.
    A command should like: `TEST 1 other information about this test`,
    and will return 1.
    """
    if len(agent.commands) == 0:  # pragma: no cover
        return 0
    command = agent.commands[0]
    if command.startswith("TEST"):
        agent.commands = agent.commands[1:]
        return int(command.split(" ")[1])
    return 0


def enable_logging():  # pragma: no cover
    import logging

    logging.basicConfig(level=logging.INFO)


def get_pidx_cidx(cmd):
    return int(cmd[2][1]), int(cmd[2][3])


def check_usage(data, usage_strs: List[str]):
    usage = [int(x) for x in usage_strs]
    assert len(data) == len(usage)
    for u, d in zip(usage, data):
        assert u == d.usage


def read_from_log_json(path: str) -> Tuple[Match, InteractionAgent, InteractionAgent]:
    """
    read from log json file, return Match instance and two agents.
    """
    data = json.load(open(path, "r", encoding="utf-8"))
    decks = [Deck.from_str(data["start_deck"][0]), Deck.from_str(data["start_deck"][1])]
    config = MatchConfig(**data["match_config"])
    random_state = data["match_random_state"]
    commands = data["command_history"]
    if isinstance(commands[0][0], (list, tuple)):  # pragma: no cover
        # new version, extract only string
        commands = [[y[1] for y in x] for x in commands]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=commands[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=commands[1], only_use_command=True
    )
    match = Match(random_state=random_state, config=config)
    match.set_deck(decks)
    return match, agent_0, agent_1


def do_log_tests(
    json_path,
    hp_modify: int = -1,
    omnipotent: bool = True,
    other_tests: Dict[int, Callable] = {},
):
    """
    Do tests on log data generated by HTTPServer, and modified by adding test commands
    with test template.
    """
    match, agent_0, agent_1 = read_from_log_json(json_path)
    match.config.history_level = 0
    # modify hp
    if hp_modify > 0:
        for i in range(2):
            characters = match.player_tables[i].player_deck_information.characters  # noqa: E501
            for c in characters:
                c.hp = c.max_hp = hp_modify
    # add omnipotent guide
    if omnipotent:
        match.event_handlers.append(OmnipotentGuideEventHandler_3_3())
    match.start()
    match.step()
    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                hps = cmd[2:]
                hps = [int(x) for x in hps]
                half = len(hps) // 2
                hps = [hps[:half], hps[half:]]
                check_hp(match, hps)
            elif test_id == 2:
                pidx = int(cmd[2][1])
                check_usage(match.player_tables[pidx].summons, cmd[3:])
            elif test_id == 3:
                pidx = int(cmd[2][1])
                check_usage(match.player_tables[pidx].supports, cmd[3:])
            elif test_id == 4:
                pidx = int(cmd[2][1])
                check_usage(match.player_tables[pidx].team_status, cmd[3:])
            elif test_id == 5:
                pidx, cidx = get_pidx_cidx(cmd)
                check_usage(match.player_tables[pidx].characters[cidx].status, cmd[3:])
            elif test_id == 6:
                pidx = int(cmd[2][1])
                assert len(match.player_tables[pidx].hands) == int(cmd[3])
            elif test_id == 7:
                pidx = int(cmd[2][1])
                d = {}
                for c in match.player_tables[pidx].dice.colors:
                    d[c.value] = d.get(c.value, 0) + 1
                for c in cmd[3:]:
                    d[c] -= 1
                    if d[c] == 0:
                        del d[c]
                assert len(d) == 0
            elif test_id == 8:
                charges = []
                for table in match.player_tables:
                    for c in table.characters:
                        charges.append(c.charge)
                assert charges == [int(x) for x in cmd[2:]]
            elif test_id == 9:
                pidx = int(cmd[2][1])
                assert len(match.player_tables[pidx].summons) == len(cmd[3:])
                for i, s in enumerate(match.player_tables[pidx].summons):
                    assert s.damage_elemental_type.value == cmd[3 + i]
            elif test_id == 10:
                pidx, cidx = get_pidx_cidx(cmd)
                desc = match.player_tables[pidx].characters[cidx].desc
                cmd_desc = " ".join(cmd[3:])[1:-1]
                assert desc == cmd_desc
            elif test_id == 11:
                card_idxs = [int(x) for x in cmd[2:]]
                card_idxs.sort()
                current = []
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        current.append(req.card_idx)
                current.sort()
                assert current == card_idxs
            elif test_id == 12:
                pidx, cidx = get_pidx_cidx(cmd)
                ele_app = match.player_tables[pidx].characters[cidx].element_application
                now_ele = [x.value for x in ele_app]
                assert now_ele == cmd[3:]
            else:
                if test_id in other_tests:
                    other_tests[test_id](match, cmd)
                else:
                    raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR



from typing import Any, Dict, List, Literal, Tuple

from ....utils.class_registry import register_base_class, register_class
from ...dice import Dice

from ...action import (
    Actions,
    CreateDiceAction,
    DrawCardAction,
    GenerateRerollDiceRequestAction,
    MakeDamageAction,
    RemoveDiceAction,
    RemoveObjectAction,
)
from ...event import (
    MoveObjectEventArguments,
    PlayerActionStartEventArguments,
    SkillEndEventArguments,
    UseCardEventArguments,
)
from ...modifiable_values import (
    CostValue,
    DamageValue,
    InitialDiceColorValue,
    RerollValue,
)
from ...event import RoundEndEventArguments, RoundPrepareEventArguments

from ...struct import Cost
from ...consts import (
    DIE_COLOR_TO_ELEMENT,
    ELEMENT_DEFAULT_ORDER,
    ELEMENT_TO_DIE_COLOR,
    CostLabels,
    DamageElementalType,
    DamageType,
    DieColor,
    IconType,
    ObjectPositionType,
)
from .base import (
    RoundEffectSupportBase,
    SupportBase,
    UsageWithRoundRestrictionSupportBase,
)


class LocationBase(SupportBase):
    cost_label: int = CostLabels.CARD.value | CostLabels.LOCATION.value


register_base_class(LocationBase)


class RoundEffectLocationBase(RoundEffectSupportBase, LocationBase):
    cost_label: int = CostLabels.CARD.value | CostLabels.LOCATION.value


class LiyueHarborWharf_3_3(LocationBase):
    name: Literal["Liyue Harbor Wharf"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(same_dice_number=2)
    usage: int = 2
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[DrawCardAction | RemoveObjectAction]:
        """
        When in round end, draw 2 cards, and check if should remove.
        """
        if self.position.area != "SUPPORT":
            # not in support area, do nothing
            return []
        self.usage -= 1
        return [
            DrawCardAction(
                player_idx=self.position.player_idx,
                number=2,
                draw_if_filtered_not_enough=True,
            ),
        ] + self.check_should_remove()


class KnightsOfFavoniusLibrary_3_3(LocationBase):
    name: Literal["Knights of Favonius Library"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(same_dice_number=1)
    usage: int = 0
    icon_type: Literal[IconType.NONE] = IconType.NONE

    def play(self, match: Any) -> List[GenerateRerollDiceRequestAction]:
        return [
            GenerateRerollDiceRequestAction(
                player_idx=self.position.player_idx,
                reroll_times=1,
            )
        ]

    def value_modifier_REROLL(
        self, value: RerollValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> RerollValue:
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return value
        if value.player_idx != self.position.player_idx:
            # not self player
            return value
        value.value += 1
        return value


class JadeChamber_4_0(LocationBase):
    name: Literal["Jade Chamber"]
    version: Literal["4.0"] = "4.0"
    cost: Cost = Cost()
    usage: int = 0
    icon_type: Literal[IconType.NONE] = IconType.NONE

    def value_modifier_INITIAL_DICE_COLOR(
        self, value: InitialDiceColorValue, match: Any, mode: Literal["REAL", "TEST"]
    ) -> InitialDiceColorValue:
        """
        If self in support, fix two dice colors to self
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return value
        if value.position.player_idx != self.position.player_idx:
            # not self player
            return value
        active = match.player_tables[self.position.player_idx].get_active_character()
        element = active.element
        value.dice_colors += [ELEMENT_TO_DIE_COLOR[element]] * 2
        return value


class JadeChamber_3_3(JadeChamber_4_0):
    version: Literal["3.3"]
    cost: Cost = Cost(same_dice_number=1)


class DawnWinery_3_3(RoundEffectLocationBase):
    name: Literal["Dawn Winery"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(same_dice_number=2)
    max_usage_per_round: int = 1

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return value
        if value.position.player_idx != self.position.player_idx:
            # not self player
            return value
        if self.usage <= 0:
            # no usage
            return value
        if value.cost.label & CostLabels.SWITCH_CHARACTER.value == 0:
            # not switch character
            return value
        # decrease cost
        if value.cost.decrease_cost(None):
            if mode == "REAL":
                self.usage -= 1
        return value


class WangshuInn_3_3(LocationBase):
    name: Literal["Wangshu Inn"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(same_dice_number=2)
    usage: int = 2
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | RemoveObjectAction]:
        """
        If back character not full hp, heal 2 hp, and check if should remove.
        """
        if self.position.area != "SUPPORT":
            # not in support area, do nothing
            return []
        table = match.player_tables[self.position.player_idx]
        characters = table.characters
        active = table.active_character_idx
        character = None
        for cnum, c in enumerate(characters):
            if cnum == active or c.is_defeated:
                # active or defeated, do nothing
                continue
            if character is None or c.damage_taken > character.damage_taken:
                character = c
        if character is None:
            # no character on standby, do nothing
            return []
        if character.damage_taken <= 0:
            # full hp, do nothing
            return []
        self.usage -= 1
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.HEAL,
                        target_position=character.position,
                        damage=-2,
                        damage_elemental_type=DamageElementalType.HEAL,
                        cost=self.cost.copy(),
                    )
                ],
            )
        ] + self.check_should_remove()


class FavoniusCathedral_3_3(LocationBase):
    name: Literal["Favonius Cathedral"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(same_dice_number=2)
    usage: int = 2
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | RemoveObjectAction]:
        """
        If active character not full hp, heal 2 hp, and check if should remove.
        """
        if self.position.area != "SUPPORT":
            # not in support area, do nothing
            return []
        character = match.player_tables[self.position.player_idx].get_active_character()
        if character.damage_taken <= 0:
            # full hp, do nothing
            return []
        self.usage -= 1
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.HEAL,
                        target_position=character.position,
                        damage=-2,
                        damage_elemental_type=DamageElementalType.HEAL,
                        cost=self.cost.copy(),
                    )
                ],
            )
        ] + self.check_should_remove()


class Tenshukaku_3_7(LocationBase):
    name: Literal["Tenshukaku"]
    version: Literal["3.7"] = "3.7"
    cost: Cost = Cost(same_dice_number=2)
    usage: int = 0
    icon_type: Literal[IconType.NONE] = IconType.NONE

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[CreateDiceAction]:
        """
        When in round prepare, if have 5 different kinds of elemental dice,
        create 1 omni element.
        """
        if self.position.area != "SUPPORT":
            # not in support area, do nothing
            return []
        colors = match.player_tables[self.position.player_idx].dice.colors
        omni_number = 0
        color_set = set()
        for color in colors:
            if color == DieColor.OMNI:
                omni_number += 1
            else:
                color_set.add(color)
        if len(color_set) + omni_number < 5:
            # not enough different kinds of elemental dice
            return []
        return [
            CreateDiceAction(
                player_idx=self.position.player_idx, number=1, color=DieColor.OMNI
            )
        ]


class GrandNarukamiShrine_3_6(LocationBase):
    name: Literal["Grand Narukami Shrine"]
    version: Literal["3.6"] = "3.6"
    cost: Cost = Cost(same_dice_number=2)
    usage: int = 3
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def play(self, match: Any) -> List[Actions]:
        """
        When played, set usage to 2 and generate one random die.
        """
        self.usage = 2
        return [
            CreateDiceAction(
                player_idx=self.position.player_idx,
                number=1,
                different=True,  # use different to avoid OMNI die
            )
        ]

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[CreateDiceAction | RemoveObjectAction]:
        """
        create one random die, and out of usage remove self.
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        self.usage -= 1
        return [
            CreateDiceAction(
                player_idx=self.position.player_idx,
                number=1,
                different=True,  # use different to avoid OMNI die
            )
        ] + self.check_should_remove()


class SangonomiyaShrine_3_7(LocationBase):
    name: Literal["Sangonomiya Shrine"]
    version: Literal["3.7"] = "3.7"
    cost: Cost = Cost(same_dice_number=2)
    usage: int = 2
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | RemoveObjectAction]:
        """
        If active character not full hp, heal 2 hp, and check if should remove.
        """
        if self.position.area != "SUPPORT":
            # not in support area, do nothing
            return []
        characters = match.player_tables[self.position.player_idx].characters
        damage_action = MakeDamageAction(
            damage_value_list=[],
        )
        for character in characters:
            if character.is_defeated or character.damage_taken <= 0:
                # full hp or defeated, do nothing
                continue
            damage_action.damage_value_list.append(
                DamageValue(
                    position=self.position,
                    damage_type=DamageType.HEAL,
                    target_position=character.position,
                    damage=-1,
                    damage_elemental_type=DamageElementalType.HEAL,
                    cost=self.cost.copy(),
                )
            )
        if len(damage_action.damage_value_list) == 0:
            # no character to heal, do nothing
            return []
        self.usage -= 1
        return [damage_action] + self.check_should_remove()


class SumeruCity_3_7(RoundEffectLocationBase):
    name: Literal["Sumeru City"]
    version: Literal["3.7"] = "3.7"
    cost: Cost = Cost(same_dice_number=2)
    max_usage_per_round: int = 1

    def play(self, match: Any) -> List[Actions]:
        """
        When played, reset usage.
        """
        self.usage = 1
        return super().play(match)

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        """
        When self character use skills or equip talents, and have usage,
        and have less or equal elemental dice than cards in hand, reduce cost.
        """
        if self.position.area != "SUPPORT":
            # not in support area, do nothing
            return value
        if self.usage == 0:
            # no usage
            return value
        if value.position.player_idx != self.position.player_idx:
            # not self player
            return value
        label = (
            CostLabels.NORMAL_ATTACK.value
            | CostLabels.ELEMENTAL_SKILL.value
            | CostLabels.ELEMENTAL_BURST.value
            | CostLabels.TALENT.value
        )
        if value.cost.label & label == 0:
            # not skill or talent
            return value
        table = match.player_tables[self.position.player_idx]
        if len(table.dice.colors) > len(table.hands):
            # more elemental dice than cards in hand
            return value
        # reduce cost
        if value.cost.decrease_cost(value.cost.elemental_dice_color):
            # success
            if mode == "REAL":
                self.usage -= 1
        return value


class Vanarana_3_7(LocationBase):
    name: Literal["Vanarana"]
    version: Literal["3.7"] = "3.7"
    cost: Cost = Cost()
    usage: int = 0
    colors: List[DieColor] = []
    icon_type: Literal[IconType.COUNTER] = IconType.COUNTER

    def play(self, match: Any) -> List[Actions]:
        self.usage = 0
        self.colors = []
        return super().play(match)

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[CreateDiceAction]:
        """
        When in round prepare, give collected dice back.
        """
        assert self.usage == len(self.colors)
        if self.position.area != "SUPPORT":
            # not in support area, do nothing
            return []
        if self.usage == 0:
            # no dice, do nothing
            return []
        ret: List[CreateDiceAction] = []
        for color in self.colors:
            ret.append(
                CreateDiceAction(
                    player_idx=self.position.player_idx, color=color, number=1
                )
            )
        self.usage = 0
        self.colors = []
        return ret

    def _sort_colors(self, colors: List[DieColor]) -> List[Tuple[int, int, DieColor]]:
        """
        Sort colors. Omni always last, others has two keys, 1. number of
        this color, 2. default order.
        Return: List[Tuple[number, default_order, color]]. For Omni, the
        default order is filled with 0.
        """
        colors_dict: Dict[DieColor, int] = {DieColor.OMNI: 0}
        for element in ELEMENT_DEFAULT_ORDER:
            colors_dict[ELEMENT_TO_DIE_COLOR[element]] = 0
        for color in colors:
            colors_dict[color] += 1
        # first: number, second: default order, third: color
        colors_lists: List[Tuple[int, int, DieColor]] = []
        # specially handle OMNI
        omni_num = 0
        for color in colors_dict:
            if colors_dict[color] == 0:
                continue
            if color == DieColor.OMNI:
                omni_num = colors_dict[color]
                continue
            colors_lists.append(
                (
                    colors_dict[color],
                    ELEMENT_DEFAULT_ORDER.index(DIE_COLOR_TO_ELEMENT[color]),
                    color,
                )
            )
        # number desc, default order asc
        colors_lists.sort(key=lambda x: (-x[0], x[1]))
        if omni_num > 0:
            colors_lists.append((omni_num, 0, DieColor.OMNI))
        return colors_lists

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[RemoveDiceAction]:
        """
        When in round end, collect up to 2 unused elemental dice.
        """
        if self.position.area != "SUPPORT":
            # not in support area, do nothing
            return []
        assert self.usage == 0 and len(self.colors) == 0
        current_dice: Dice = match.player_tables[self.position.player_idx].dice
        current_dice_colors = current_dice.colors
        if len(current_dice_colors) == 0:
            # no dice, do nothing
            return []
        colors_lists = self._sort_colors(current_dice_colors)
        assert len(colors_lists) > 0
        # if first has more than 2, take two of them
        if colors_lists[0][0] >= 2:
            self.colors = [colors_lists[0][2]] * 2
        else:
            # otherwise, take first one
            self.colors = [colors_lists[0][2]]
            # if has second, take second one
            if len(colors_lists) >= 2:
                self.colors += [colors_lists[1][2]]
        self.usage = len(self.colors)
        dice_idxs = current_dice.colors_to_idx(self.colors)
        return [
            RemoveDiceAction(
                player_idx=self.position.player_idx,
                dice_idxs=dice_idxs,
            )
        ]


class ChinjuForest_3_7(LocationBase):
    name: Literal["Chinju Forest"]
    version: Literal["3.7"] = "3.7"
    cost: Cost = Cost(same_dice_number=1)
    usage: int = 3
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[CreateDiceAction | RemoveObjectAction]:
        """
        When in round prepare, if have usage, create one die that match
        active character element.
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        if self.usage <= 0:  # pragma: no cover
            # no usage
            return []
        if event.player_go_first == self.position.player_idx:
            # self go first, do nothing
            return []
        table = match.player_tables[self.position.player_idx]
        active = table.get_active_character()
        element = active.element
        self.usage -= 1
        return [
            CreateDiceAction(
                player_idx=self.position.player_idx,
                number=1,
                color=ELEMENT_TO_DIE_COLOR[element],
            )
        ] + self.check_should_remove()


class GoldenHouse_4_0(LocationBase, UsageWithRoundRestrictionSupportBase):
    name: Literal["Golden House"]
    version: Literal["4.0"] = "4.0"
    cost: Cost = Cost()
    usage: int = 2
    max_usage_one_round: int = 1
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    card_cost_label: int = CostLabels.WEAPON.value
    decrease_threshold: int = 3

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        """
        if card label match and original cost greater than threshold,
        reduce cost.
        """
        assert value.cost.original_value is not None
        if (
            self.position.area == ObjectPositionType.SUPPORT
            and value.position.player_idx == self.position.player_idx
            and value.cost.original_value.total_dice_cost >= self.decrease_threshold
            and value.cost.label & self.card_cost_label != 0
            and self.has_usage()
        ):
            # area right, player right, cost label match, cost greater than
            # threshold, and not used this round
            if value.cost.decrease_cost(None):
                if mode == "REAL":
                    self.use()
        return value

    def event_handler_MOVE_OBJECT(
        self, event: MoveObjectEventArguments, match: Any
    ) -> List[Actions]:
        if self.usage == 0:
            # if usage is zero, it run out of usage, remove it
            return list(self.check_should_remove())
        # otherwise, do parent event handler
        return super().event_handler_MOVE_OBJECT(event, match)


class GandharvaVille_4_1(LocationBase, UsageWithRoundRestrictionSupportBase):
    name: Literal["Gandharva Ville"]
    version: Literal["4.1"] = "4.1"
    cost: Cost = Cost(same_dice_number=1)
    usage: int = 3
    max_usage_one_round: int = 1
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def event_handler_PLAYER_ACTION_START(
        self, event: PlayerActionStartEventArguments, match: Any
    ) -> List[Actions]:
        """
        If self action start, and has usage, and no elemental dice, create
        one omni element.
        """
        if not (
            self.position.area == ObjectPositionType.SUPPORT
            and event.player_idx == self.position.player_idx
            and self.has_usage()
            and len(match.player_tables[self.position.player_idx].dice.colors) == 0
        ):
            # not equipped, or not self, or no usage, or have elemental dice
            return []
        # create one omni element
        self.use()
        return [
            CreateDiceAction(
                player_idx=self.position.player_idx, number=1, color=DieColor.OMNI
            )
        ]


class StormterrorsLair_4_2(LocationBase, UsageWithRoundRestrictionSupportBase):
    name: Literal["Stormterror's Lair"]
    version: Literal["4.2"] = "4.2"
    cost: Cost = Cost(same_dice_number=2)
    usage: int = 3
    max_usage_one_round: int = 1
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    card_cost_label: int = CostLabels.SKILLS.value
    decrease_threshold: int = 4

    def play(self, match: Any) -> List[Actions]:
        """
        Draw a talent card
        """
        return super().play(match) + [
            DrawCardAction(
                player_idx=self.position.player_idx,
                number=1,
                whitelist_cost_labels=CostLabels.TALENT.value,
                draw_if_filtered_not_enough=False,
            )
        ]

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        """
        if card label match and original cost greater than threshold,
        or is talent cost, reduce cost.
        """
        if (
            self.position.area == ObjectPositionType.SUPPORT
            and value.position.player_idx == self.position.player_idx
            and self.has_usage()
        ):
            # area right, player right, and not used this round
            assert value.cost.original_value is not None
            if (
                value.cost.original_value.total_dice_cost >= self.decrease_threshold
                and value.cost.label & self.card_cost_label != 0
            ) or value.cost.label & CostLabels.TALENT.value != 0:
                # cost label match, cost greater than threshold; or is talent
                if value.cost.decrease_cost(value.cost.elemental_dice_color):
                    if mode == "REAL":
                        self.use()
        return value

    def event_handler_USE_CARD(
        self, event: UseCardEventArguments, match: Any
    ) -> List[Actions]:
        if self.usage == 0:
            # if usage is zero, it run out of usage, remove it
            return list(self.check_should_remove())
        # otherwise, do parent event handler
        return super().event_handler_USE_CARD(event, match)

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        if self.usage == 0:
            # if usage is zero, it run out of usage, remove it
            return list(self.check_should_remove())
        return []


register_class(
    LiyueHarborWharf_3_3
    | KnightsOfFavoniusLibrary_3_3
    | JadeChamber_4_0
    | JadeChamber_3_3
    | DawnWinery_3_3
    | WangshuInn_3_3
    | FavoniusCathedral_3_3
    | Tenshukaku_3_7
    | GrandNarukamiShrine_3_6
    | SangonomiyaShrine_3_7
    | SumeruCity_3_7
    | Vanarana_3_7
    | ChinjuForest_3_7
    | GoldenHouse_4_0
    | GandharvaVille_4_1
    | StormterrorsLair_4_2
)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...action import RemoveObjectAction

from ...event import MoveObjectEventArguments, UseSkillEventArguments

from ...consts import CostLabels, IconType

from ...modifiable_values import CostValue
from .base import RoundCharacterStatus, ShieldCharacterStatus


class LithicSpear_3_3(ShieldCharacterStatus):
    name: Literal["Lithic Spear"] = "Lithic Spear"
    version: Literal["3.3"] = "3.3"
    usage: int
    max_usage: int

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def renew(self, object: "LithicSpear_3_3") -> None:
        self.max_usage = object.max_usage
        self.usage = max(object.usage, self.usage)
        self.usage = min(self.max_usage, self.usage)


class KingsSquire_4_0(RoundCharacterStatus):
    name: Literal["King's Squire"] = "King's Squire"
    version: Literal["4.0"] = "4.0"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.BUFF] = IconType.BUFF

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        """
        If self use elemental skill, or equip talent, cost -2.
        """
        if not self._check_value_self_skill_or_talent(
            value, match, CostLabels.ELEMENTAL_SKILL.value
        ):
            return value
        # decrease cost
        assert self.usage > 0
        decrease_result = [
            value.cost.decrease_cost(value.cost.elemental_dice_color),
            value.cost.decrease_cost(value.cost.elemental_dice_color),
        ]
        if True in decrease_result:  # pragma: no branch
            # decrease cost success
            if mode == "REAL":
                self.usage -= 1
        return value

    def event_handler_USE_SKILL(
        self, event: UseSkillEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        return self.check_should_remove()

    def event_handler_MOVE_OBJECT(
        self, event: MoveObjectEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        return self.check_should_remove()


class Moonpiercer_4_1(KingsSquire_4_0):
    name: Literal["Moonpiercer"] = "Moonpiercer"
    version: Literal["4.1"] = "4.1"


register_class(LithicSpear_3_3 | KingsSquire_4_0 | Moonpiercer_4_1)



from lpsim.agents.interaction_agent import InteractionAgent
from lpsim.server.match import Match, MatchState
from lpsim.server.deck import Deck
from tests.utils_for_test import (
    check_hp,
    get_random_state,
    get_test_id_from_command,
    make_respond,
    set_16_omni,
)


def test_elec_hypo():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "TEST 2 2 skill 0 1",
            "skill 1 0 1 2 3 4",
            "TEST 1 4 10 10 3 8 10",
            "skill 0 0 1 2",
            "skill 4 0 1 2",
            "TEST 3 sw_char cost 1",
            "sw_char 1 0",
            "end",
            "sw_char 2 0",
            "sw_char 1 0",
            "skill 0 0 1 2",
            "sw_char 0 0",
            "skill 1 0 1 2 3 4",
            "TEST 1 1 8 8 1 6 10",
            "end",
        ],
        [
            "sw_card",
            "choose 1",
            "TEST 1 8 10 10 8 8 10",
            "sw_char 0 0",
            "TEST 1 8 10 10 6 8 10",
            "skill 1 0 1 2 3 4",
            "TEST 1 1 10 10 1 8 10",
            "TEST 3 switch_character cost 2",
            "card 0 0 0 1 2",
            "sw_char 1 0 1",
            "sw_char 0 0",
            "end",
            "TEST 3 sw_char cost 2",
            "skill 1 0 1 2 3 4",
            "TEST 4 active p1",
            "TEST 5 p0c1 status empty",
            "sw_char 0 0 1",
            "sw_char 1 0",
            "skill 1 0 1 2",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Electro Hypostasis
        character:Klee
        character:Keqing
        Absorbing Prism*15
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                found = set()
                for req in match.requests:
                    if req.name == "UseSkillRequest":
                        found.add(req.skill_idx)
                assert found == set([0, 1])
            elif test_id == 3:
                cmd = cmd.split()
                sc = int(cmd[-1])
                for req in match.requests:
                    if req.name == "SwitchCharacterRequest":
                        assert sc == req.cost.total_dice_cost
            elif test_id == 4:
                assert match.current_player == 1
            elif test_id == 5:
                assert len(match.player_tables[0].characters[1].status) == 0
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_2():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "sw_char 2 0",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "end",
            "sw_char 0 0",
            "end",
            "end",
            "choose 1",
            "sw_char 2 0",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "choose 1",
        ],
        [
            "sw_card",
            "choose 1",
            "sw_char 2 0",
            "skill 0 0 1 2",
            "sw_char 0 0",
            "sw_char 1 0",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "end",
            "skill 0 0 1 2",
            "sw_char 2 0",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "end",
            "skill 0 0 1 2",
            "sw_char 1 0",
            "skill 2 0 1 2",
            "sw_char 0 0",
            "end",
            "skill 1 0 1 2 3 4",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Electro Hypostasis
        character:Klee
        character:Keqing
        Absorbing Prism*15
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_prepare_frozen():
    cmd_records = [
        ["sw_card", "choose 0", "skill 1 0 1 2 3 4", "end"],
        [
            "sw_card",
            "choose 2",
            "skill 1 0 1 2",
            "sw_char 1 0",
            "skill 0 0 1 2",
            "TEST 1 5 10 10 8 10 7",
            "end",
            "TEST 1 4 10 10 8 7 7",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Electro Hypostasis
        character:CryoMobMage
        character:Mona
        Absorbing Prism*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_electro_bennett():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "card 0 0 0 1 2 3",
            "sw_char 2 0",
            "skill 1 0 1 2 3 4",
        ],
        [
            "sw_card",
            "choose 0",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "sw_char 1 0",
            "skill 0 0 1 2",
            "TEST 1 8 10 8 4 6 8",
            "skill 0 0 1 2",
            "sw_char 2 0",
            "TEST 1 8 10 6 4 2 3",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Bennett
        character:Maguu Kenki
        character:Electro Hypostasis
        Grand Expectation*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_frozen_talent():
    cmd_records = [
        ["sw_card", "choose 2", "skill 1 15 14 13", "sw_char 1 12", "skill 1 11 10 9"],
        ["sw_card", "choose 0", "skill 0 15 14 13", "card 0 0 12 11"],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.2
        character:Electro Hypostasis
        character:Kaeya
        character:Mona
        Absorbing Prism*15
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    # test_elec_hypo()
    # test_prepare_frozen()
    # test_electro_bennett()
    test_frozen_talent()



from typing import Any, List, Literal, Tuple

from ....utils.class_registry import register_class

from ...summon.base import AttackerSummonBase

from ...event import (
    ChooseCharacterEventArguments,
    RoundPrepareEventArguments,
    SwitchCharacterEventArguments,
)

from ...action import Actions, CreateObjectAction, RemoveObjectAction
from ...struct import Cost, ObjectPosition

from ...consts import (
    CostLabels,
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    PlayerActionLabels,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalNormalAttackBase,
    ElementalSkillBase,
    CharacterBase,
    TalentBase,
)


# Summons


class LightningRoseSummon_4_0(AttackerSummonBase):
    name: Literal["Lightning Rose"] = "Lightning Rose"
    version: Literal["4.0"] = "4.0"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.ELECTRO
    damage: int = 2


# Skills


class LightningTouch(ElementalNormalAttackBase):
    name: Literal["Lightning Touch"] = "Lightning Touch"
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = ElementalNormalAttackBase.get_cost(ElementType.ELECTRO)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Do attack; if charged attack, attach conductive
        """
        if not match.player_tables[self.position.player_idx].charge_satisfied:
            # not charged attack
            return super().get_actions(match)
        return super().get_actions(match) + [
            self.create_opposite_character_status(match, "Conductive", {})
        ]


class VioletArc(ElementalSkillBase):
    name: Literal["Violet Arc"] = "Violet Arc"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        If Conductive not attached, attach Conductive
        otherwise, increase damage.
        """
        target = match.player_tables[
            1 - self.position.player_idx
        ].get_active_character()
        status = target.status
        conductive = None
        for s in status:
            if s.name == "Conductive":
                conductive = s
                break
        if conductive is None:
            # no conductive
            return super().get_actions(match) + [
                self.create_opposite_character_status(match, "Conductive", {})
            ]
        else:
            # has conductive
            self.damage += conductive.usage
            ret = super().get_actions(match)
            self.damage = 2
            ret = [RemoveObjectAction(object_position=conductive.position)] + ret
            return ret


class LightningRose(ElementalBurstBase):
    name: Literal["Lightning Rose"] = "Lightning Rose"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [self.create_summon(self.name)]


# Talents


class PulsatingWitch_4_0(TalentBase):
    name: Literal["Pulsating Witch"]
    version: Literal["4.0"] = "4.0"
    character_name: Literal["Lisa"] = "Lisa"
    cost: Cost = Cost(elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=1)
    cost_label: int = (
        CostLabels.CARD.value | CostLabels.TALENT.value | CostLabels.EQUIPMENT.value
    )
    usage: int = 1

    def is_valid(self, match: Any) -> bool:
        """
        Only corresponding character is active character can equip this card.
        """
        if self.position.area != ObjectPositionType.HAND:
            # not in hand, cannot equip
            raise AssertionError("Talent is not in hand")
        return len(self.get_targets(match)) > 0

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        All ayaka are targets
        """
        assert self.position.area != ObjectPositionType.CHARACTER
        ret: List[ObjectPosition] = []
        for c in match.player_tables[self.position.player_idx].characters:
            if c.name == self.character_name and c.is_alive:
                ret.append(c.position)
        return ret

    def get_action_type(self, match: Any) -> Tuple[int, bool]:
        return (PlayerActionLabels.CARD.value | PlayerActionLabels.SKILL.value, False)

    def _attach_conductive_to_opposite_active(
        self,
        event: SwitchCharacterEventArguments | ChooseCharacterEventArguments,
        match: Any,
    ) -> List[CreateObjectAction]:
        if (
            event.action.player_idx != self.position.player_idx
            or event.action.character_idx != self.position.character_idx
        ):
            # not to this character
            return []
        target = match.player_tables[
            1 - self.position.player_idx
        ].get_active_character()
        if self.usage <= 0:
            # no usage
            return []
        self.usage -= 1
        return [
            CreateObjectAction(
                object_name="Conductive",
                object_position=target.position.set_area(
                    ObjectPositionType.CHARACTER_STATUS
                ),
                object_arguments={},
            )
        ]

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        Reset usage
        """
        self.usage = 1
        return []

    def event_handler_SWITCH_CHARACTER(
        self, event: SwitchCharacterEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        return self._attach_conductive_to_opposite_active(event, match)

    def event_handler_CHOOSE_CHARACTER(
        self, event: ChooseCharacterEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        return self._attach_conductive_to_opposite_active(event, match)


# character base


class Lisa_4_0(CharacterBase):
    name: Literal["Lisa"]
    version: Literal["4.0"] = "4.0"
    element: ElementType = ElementType.ELECTRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[LightningTouch | VioletArc | LightningRose] = []
    faction: List[FactionType] = [FactionType.MONDSTADT]
    weapon_type: WeaponType = WeaponType.CATALYST

    def _init_skills(self) -> None:
        self.skills = [LightningTouch(), VioletArc(), LightningRose()]


register_class(Lisa_4_0 | PulsatingWitch_4_0 | LightningRoseSummon_4_0)



from enum import Enum
from typing import List, Any
from ..utils import BaseModel
from .consts import (
    DieColor,
    DamageType,
    DamageElementalType,
    ElementalReactionType,
    ElementType,
    ObjectPositionType,
    SkillType,
)
from .struct import ObjectPosition, Cost


class ModifiableValueTypes(str, Enum):
    """
    Enum representing the type of a modifiable value. A modifiable value is a
    value that can be modified by other objects before being used.
    """

    INITIAL_DICE_COLOR = "INITIAL_DICE_COLOR"
    REROLL = "REROLL"
    COMBAT_ACTION = "COMBAT_ACTION"

    # Cost is used in most situations that decrease cost. Full cost is used
    # only for effects that should decrease cost to 0, which is performed after
    # normal cost decrease. Until 4.2, only Timaeus and Wagner has this effect.
    COST = "COST"
    FULL_COST = "FULL_COST"

    # damage calculation is split into 4 parts: increase, multiply and
    # decrease. damage will be first increased, then multiplied,
    # then decreased.
    DAMAGE_ELEMENT_ENHANCE = "DAMAGE_ELEMENT_ENHANCE"
    DAMAGE_INCREASE = "DAMAGE_INCREASE"
    DAMAGE_MULTIPLY = "DAMAGE_MULTIPLY"
    DAMAGE_DECREASE = "DAMAGE_DECREASE"

    # before using card, a modifiable value is used to check whether use card
    # successfully.
    USE_CARD = "USE_CARD"


class ModifiableValueBase(BaseModel):
    """
    Base class of modifiable values. It saves the value and can pass through
    value modifiers to update its value in-place.
    """

    type: ModifiableValueTypes
    position: ObjectPosition


class InitialDiceColorValue(ModifiableValueBase):
    type: ModifiableValueTypes = ModifiableValueTypes.INITIAL_DICE_COLOR
    player_idx: int
    dice_colors: List[DieColor] = []


class RerollValue(ModifiableValueBase):
    type: ModifiableValueTypes = ModifiableValueTypes.REROLL
    value: int
    player_idx: int


class CostValue(ModifiableValueBase):
    type: ModifiableValueTypes = ModifiableValueTypes.COST
    target_position: ObjectPosition | None
    cost: Cost

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        self.cost = self.cost.copy()
        if self.cost.original_value is None:  # pragma: no branch
            self.cost.original_value = self.cost.copy()


class FullCostValue(CostValue):
    type: ModifiableValueTypes = ModifiableValueTypes.FULL_COST

    def __init__(self, *argv, **kwargs):
        # full cost is used after normal cost decrease, so no need to refresh
        # its original value, just copy normal cost value.
        BaseModel.__init__(self, *argv, **kwargs)

    @staticmethod
    def from_cost_value(value: CostValue) -> "FullCostValue":
        return FullCostValue(
            position=value.position,
            target_position=value.target_position,
            cost=value.cost,
        )


class CombatActionValue(ModifiableValueBase):
    type: ModifiableValueTypes = ModifiableValueTypes.COMBAT_ACTION
    action_label: int
    do_combat_action: bool


class DamageElementEnhanceValue(ModifiableValueBase):
    """
    It describes a detailed damage, i.e. character X will receive the damage.
    It contains all information that a damage is needed, except element
    reaction, as it is not decided yet.

    It is the initial stage of damage calculation.
    In this stage, buffs will modify the damage elemental type.
    """

    type: ModifiableValueTypes = ModifiableValueTypes.DAMAGE_ELEMENT_ENHANCE
    damage_type: DamageType
    target_position: ObjectPosition
    damage: int
    damage_elemental_type: DamageElementalType
    cost: Cost  # original cost of source

    damage_from_element_reaction: bool = False

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        if self.damage_type == DamageType.HEAL:
            assert self.damage < 0, "Heal should be negative"
            assert (
                self.damage_elemental_type == DamageElementalType.HEAL
            ), "Heal should be heal type"
        elif self.damage_type == DamageType.DAMAGE:
            assert self.damage >= 0, "Damage should be non-negative"
        else:
            assert self.damage_type == DamageType.ELEMENT_APPLICATION
            assert self.damage == 0, "Element application should be 0"

    def copy(self, *argv, **kwargs) -> "DamageElementEnhanceValue":
        assert (
            len(argv) == 0 and len(kwargs) == 0
        ), "No parameters are allowed when copying."
        return DamageElementEnhanceValue(
            position=self.position,
            damage_type=self.damage_type,
            target_position=self.target_position,
            damage=self.damage,
            damage_elemental_type=self.damage_elemental_type,
            cost=self.cost.copy(),
            damage_from_element_reaction=self.damage_from_element_reaction,
        )

    def is_corresponding_character_use_damage_skill(
        self,
        object_position: ObjectPosition,
        match: Any,
        skill_type: SkillType | None,
        ignore_elemental_reaction: bool = True,
    ) -> bool:
        """
        Most objects modifying damage when current character
        makes damages by using skills. This function checks if the object
        with given object_position is the corresponding character using skill.

        If object position is on certain character, i.e. equipment or
        character status, the character should match damage source. Otherwise,
        the active character should be the damage source.

        Args:
            object_position: The position of the object to check.
            match: The match object.
            skill_type: If not None, the skill type should match.
            ignore_elemental_reaction: If True, elemental reaction damages will
                not be considered.
        """
        if self.damage_type != DamageType.DAMAGE:
            # not damage
            return False
        if self.position.area != ObjectPositionType.SKILL:
            # not skill
            return False
        if self.position.player_idx != object_position.player_idx:
            # not same player
            return False
        if (
            object_position.area == ObjectPositionType.CHARACTER
            or object_position.area == ObjectPositionType.CHARACTER_STATUS
        ):
            # object on character, get idx
            character_idx = object_position.character_idx
        else:
            # not on character, get active character idx
            character_idx = match.player_tables[
                object_position.player_idx
            ].active_character_idx
        if self.position.character_idx != character_idx:
            # not same character
            return False
        if ignore_elemental_reaction and self.damage_from_element_reaction:
            # elemental reaction damage
            return False
        if skill_type is not None:
            skill = match.get_object(self.position)
            if skill.skill_type != skill_type:
                # skill type not match
                return False
        return True

    def is_corresponding_character_receive_damage(
        self, object_position: ObjectPosition, match: Any, ignore_piercing: bool = True
    ) -> bool:
        """
        Most objects modifying damage only when current character
        receives damages. This function checks if the object with given
        object_position is the corresponding character receiving damage.

        If object position is on certain character, i.e. equipment or
        character status, the character should match damage target. Otherwise,
        the active character should be the damage target.

        Args:
            object_position: The position of the object to check.
            match: The match object.
            ignore_piercing: If True, piercing damage will result in False.
        """
        if self.damage_type != DamageType.DAMAGE:
            # not damage
            return False
        if (
            ignore_piercing
            and self.damage_elemental_type == DamageElementalType.PIERCING
        ):
            # piercing damage
            return False
        if (
            self.target_position.area != ObjectPositionType.CHARACTER
        ):  # pragma: no cover
            # not on character
            return False
        if self.target_position.player_idx != object_position.player_idx:
            # not same player
            return False
        if (
            object_position.area == ObjectPositionType.CHARACTER
            or object_position.area == ObjectPositionType.CHARACTER_STATUS
        ):
            # object on character, get idx
            character_idx = object_position.character_idx
        else:
            # not on character, get active character idx
            character_idx = match.player_tables[
                object_position.player_idx
            ].active_character_idx
        if self.target_position.character_idx != character_idx:
            # not same character
            return False
        return True

    @classmethod
    def create_heal(
        cls,
        source_position: ObjectPosition,
        target_position: ObjectPosition,
        heal: int,
        cost: Cost,
    ):
        """
        class method to create heal-type damage
        """
        assert heal < 0, "Heal should be negative"
        return cls(
            position=source_position,
            target_position=target_position,
            damage_type=DamageType.HEAL,
            damage=heal,
            damage_elemental_type=DamageElementalType.HEAL,
            cost=cost,
        )

    @classmethod
    def create_element_application(
        cls,
        source_position: ObjectPosition,
        target_position: ObjectPosition,
        element: DamageElementalType,
        cost: Cost,
    ):
        """
        class method to create element-application-type damage
        """
        return cls(
            position=source_position,
            target_position=target_position,
            damage_type=DamageType.ELEMENT_APPLICATION,
            damage=0,
            damage_elemental_type=element,
            cost=cost,
        )


class DamageIncreaseValue(DamageElementEnhanceValue):
    """
    It inherits DamageElementEnhanceValue, and add two information about
    elemental reaction.

    In this stage, damage elemental type has been decided, and all add-type
    damage buffs will modify the damage, including elemental reaction.
    """

    type: ModifiableValueTypes = ModifiableValueTypes.DAMAGE_INCREASE

    element_reaction: ElementalReactionType
    reacted_elements: List[ElementType]

    @staticmethod
    def from_element_enhance_value(
        value: DamageElementEnhanceValue,
        element_reaction: ElementalReactionType,
        reacted_elements: List[ElementType],
    ) -> "DamageIncreaseValue":
        return DamageIncreaseValue(
            position=value.position,
            damage_type=value.damage_type,
            target_position=value.target_position,
            damage=value.damage,
            damage_elemental_type=value.damage_elemental_type,
            cost=value.cost,
            damage_from_element_reaction=value.damage_from_element_reaction,
            element_reaction=element_reaction,
            reacted_elements=reacted_elements,
        )


class DamageMultiplyValue(DamageIncreaseValue):
    """
    In this stage, damage increase has been decided, and all multiply-type
    buffs will modify the damage, e.g. Mona elemental burst, Noelle elemental
    skill.
    """

    type: ModifiableValueTypes = ModifiableValueTypes.DAMAGE_MULTIPLY

    @staticmethod
    def from_increase_value(
        increase_value: DamageIncreaseValue,
    ) -> "DamageMultiplyValue":
        return DamageMultiplyValue(
            position=increase_value.position,
            damage_type=increase_value.damage_type,
            target_position=increase_value.target_position,
            damage=increase_value.damage,
            damage_elemental_type=increase_value.damage_elemental_type,
            damage_from_element_reaction=increase_value.damage_from_element_reaction,
            cost=increase_value.cost,
            element_reaction=increase_value.element_reaction,
            reacted_elements=increase_value.reacted_elements,
        )


class DamageDecreaseValue(DamageIncreaseValue):
    """
    In this stage, damage multiply has been decided, and all decrease-type
    buffs will modify the damage, i.e. various types of shields.
    """

    type: ModifiableValueTypes = ModifiableValueTypes.DAMAGE_DECREASE

    @staticmethod
    def from_multiply_value(
        multiply_value: DamageMultiplyValue,
    ) -> "DamageDecreaseValue":
        return DamageDecreaseValue(
            position=multiply_value.position,
            damage_type=multiply_value.damage_type,
            target_position=multiply_value.target_position,
            damage=multiply_value.damage,
            damage_elemental_type=multiply_value.damage_elemental_type,
            damage_from_element_reaction=multiply_value.damage_from_element_reaction,
            cost=multiply_value.cost,
            element_reaction=multiply_value.element_reaction,
            reacted_elements=multiply_value.reacted_elements,
        )

    def apply_shield(
        self,
        shield_usage: int,
        min_damage_to_trigger: int,
        max_in_one_time: int,
        decrease_usage_by_damage: bool,
    ) -> int:
        """
        Apply shield to damage, and return the shield usage after applying.

        Args:
            shield_usage: Current shield usage.
            decrease_usage_by_damage: If True, decrease usage by damage,
                and the following two parameters are ignored.
            min_damage_to_trigger: Minimum damage to trigger shield.
            max_in_one_time: Maximum damage to decrease in one time.
        """
        if shield_usage <= 0:
            # no usage, return
            return shield_usage
        if decrease_usage_by_damage:
            # decrease usage by damage
            decrease = min(shield_usage, self.damage)
            self.damage -= decrease
            shield_usage -= decrease
        else:
            if self.damage < min_damage_to_trigger:
                # damage too small to trigger
                return shield_usage
            decrease = min(max_in_one_time, self.damage)
            self.damage -= decrease
            shield_usage -= 1
        return shield_usage


DamageValue = DamageElementEnhanceValue
FinalDamageValue = DamageDecreaseValue  # alias


class UseCardValue(ModifiableValueBase):
    """
    Value to check whether using card success. If use_card is false, then the
    actions about the card are ignored, but will consume cost.
    """

    type: ModifiableValueTypes = ModifiableValueTypes.USE_CARD
    card: Any
    use_card: bool = True



from typing import Any, List, Literal

from ....utils.class_registry import register_class
from ...modifiable_values import CostValue, DamageIncreaseValue, DamageValue
from ...struct import Cost, ObjectPosition
from ...consts import (
    CostLabels,
    DamageElementalType,
    DamageType,
    IconType,
    ObjectPositionType,
    SkillType,
)
from ...action import (
    ActionTypes,
    Actions,
    CreateObjectAction,
    MakeDamageAction,
    RemoveObjectAction,
)
from ...event import (
    CharacterDefeatedEventArguments,
    CreateObjectEventArguments,
    MakeDamageEventArguments,
    RoundPrepareEventArguments,
    SkillEndEventArguments,
)
from .base import (
    CharacterStatusBase,
    ElementalInfusionCharacterStatus,
    PrepareCharacterStatus,
    RoundCharacterStatus,
    RoundEndAttackCharacterStatus,
    ShieldCharacterStatus,
    UsageCharacterStatus,
)


class Riptide_4_1(CharacterStatusBase):
    """
    This status will not deal damages directly, and defeat-regenerate is
    handled by system handler, which is created by Tartaglia'is passive skill
    when game starts.
    """

    name: Literal["Riptide"] = "Riptide"
    # As Riptide has changed, all status it related to should have different
    # version. To avoid using wrong version of status, related status have
    # not default value.
    version: Literal["4.1"]
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    available_handler_in_trashbin: List[ActionTypes] = [ActionTypes.CHARACTER_DEFEATED]

    def event_handler_CHARACTER_DEFEATED(
        self, event: CharacterDefeatedEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        If defeated character has Riptide, the next active character will get
        Riptide. As this event triggers later than CHOOSE_CHARACTER, current
        active character is set.
        """
        pidx = event.action.player_idx
        cidx = event.action.character_idx
        if pidx != self.position.player_idx or cidx != self.position.character_idx:
            # defeated character not self character
            return []
        return [
            CreateObjectAction(
                object_name="Riptide",
                object_position=ObjectPosition(
                    player_idx=pidx,
                    area=ObjectPositionType.CHARACTER_STATUS,
                    character_idx=match.player_tables[pidx].active_character_idx,
                    id=0,
                ),
                object_arguments={"version": self.version},
            )
        ]


class RangedStance_4_1(CharacterStatusBase):
    name: Literal["Ranged Stance"] = "Ranged Stance"
    # As Riptide has changed, all status it related to should have different
    # version. To avoid using wrong version of status, related status have
    # not default value.
    version: Literal["4.1"]
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS
    usage: int = 1
    max_usage: int = 1

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        When self character use charged attack, apply Riptide to target
        character.
        """
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        ):
            # not self character use skill
            return []
        skill = match.get_object(event.action.position)
        if (
            not match.player_tables[self.position.player_idx].charge_satisfied
            or skill.skill_type != SkillType.NORMAL_ATTACK
        ):
            # not charged attack
            return []
        character = match.player_tables[
            event.action.target_position.player_idx
        ].characters[event.action.target_position.character_idx]
        if character.is_defeated:
            # defeated character cannot attach Riptide
            return []
        return [
            CreateObjectAction(
                object_name="Riptide",
                object_position=event.action.target_position.set_area(
                    ObjectPositionType.CHARACTER_STATUS
                ),
                object_arguments={"version": self.version},
            )
        ]

    def event_handler_CREATE_OBJECT(
        self, event: CreateObjectEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        If self character gain Melee Stance, remove this status.
        """
        if (
            event.action.object_name == "Melee Stance"
            and self.position.check_position_valid(
                event.action.object_position,
                match,
                player_idx_same=True,
                character_idx_same=True,
                target_area=ObjectPositionType.CHARACTER_STATUS,
            )
        ):
            # self character gain melee stance, remove this status
            return [RemoveObjectAction(object_position=self.position)]
        return []


class MeleeStance_4_1(ElementalInfusionCharacterStatus, RoundCharacterStatus):
    name: Literal["Melee Stance"] = "Melee Stance"
    # As Riptide has changed, all status it related to should have different
    # version. To avoid using wrong version of status, related status have
    # not default value.
    version: Literal["4.1"]
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS
    usage: int = 2
    max_usage: int = 2
    infused_elemental_type: DamageElementalType = DamageElementalType.HYDRO

    extra_attack_usage: int = 2
    extra_attack_max_usage: int = 2

    def self_skill(self, position: ObjectPosition, match: Any) -> bool:
        """
        Whether position is self skill
        """
        return self.position.check_position_valid(
            position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        )

    def charge_attack(self, skill_position: ObjectPosition, match: Any) -> bool:
        """
        Whether position is charged attack
        """
        skill = match.get_object(skill_position)
        return (
            match.player_tables[skill_position.player_idx].charge_satisfied
            and skill.skill_type == SkillType.NORMAL_ATTACK
        )

    def opposite_active_has_riptide(self, match: Any) -> bool:
        """
        Whether opposite active character has Riptide
        """
        active = match.player_tables[
            1 - self.position.player_idx
        ].get_active_character()
        for status in active.status:
            if status.name == "Riptide":
                return True
        return False

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        If target has Riptide, increase damage by 1
        """
        assert mode == "REAL"
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, None
        ):
            # not this character use skill, not modify
            return value
        if not self.opposite_active_has_riptide(match):
            # opposite active character does not have Riptide
            return value
        # add damage
        value.damage += 1
        return value

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        When use skill on target with Riptide, deal 1 piercing damage to next
        character.
        """
        assert len(event.damages) > 0
        # first must be skill damage
        damage = event.damages[0]
        target_position = damage.final_damage.target_position
        if not self.self_skill(damage.final_damage.position, match):
            # not self character use skill
            return []
        if self.opposite_active_has_riptide(match):
            # opposite have Riptide, deal 1 piercing damage to next character
            next_idx = match.player_tables[
                1 - self.position.player_idx
            ].next_character_idx(target_position.character_idx)
            if next_idx is not None and self.extra_attack_usage > 0:
                # has next character and has usage:
                character = match.player_tables[
                    1 - self.position.player_idx
                ].characters[next_idx]
                self.extra_attack_usage -= 1
                return [
                    MakeDamageAction(
                        damage_value_list=[
                            DamageValue(
                                position=self.position,
                                damage_type=DamageType.DAMAGE,
                                target_position=character.position,
                                damage=1,
                                damage_elemental_type=DamageElementalType.PIERCING,
                                cost=Cost(),
                            )
                        ]
                    )
                ]
        return []

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[CreateObjectAction | MakeDamageAction]:
        """
        When self character use charged attack, apply Riptide to target
        character.
        """
        ret: List[CreateObjectAction | MakeDamageAction] = []

        if not self.self_skill(event.action.position, match):
            # not self character use skill
            return []
        # Riptide attach conditions
        if self.charge_attack(event.action.position, match):
            # charged attack
            character = match.player_tables[
                event.action.target_position.player_idx
            ].characters[event.action.target_position.character_idx]
            if character.is_alive:
                ret += [
                    CreateObjectAction(
                        object_name="Riptide",
                        object_position=event.action.target_position.set_area(
                            ObjectPositionType.CHARACTER_STATUS
                        ),
                        object_arguments={"version": self.version},
                    )
                ]

        return ret

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        Reset extra attack usage, and when this status is removed,
        attach Ranged Stance.
        """
        self.extra_attack_usage = self.extra_attack_max_usage
        ret = super().event_handler_ROUND_PREPARE(event, match)
        if len(ret) > 0:
            # has remove action
            assert ret[0].type == ActionTypes.REMOVE_OBJECT
            ret.append(
                CreateObjectAction(
                    object_name="Ranged Stance",
                    object_position=self.position,
                    object_arguments={"version": self.version},
                )
            )
        return ret


class Riptide_3_7(Riptide_4_1, RoundCharacterStatus):
    """
    This status will not deal damages directly, and defeat-regenerate is
    handled by system handler, which is created by Tartaglia'is passive skill
    when game starts.
    """

    name: Literal["Riptide"] = "Riptide"
    version: Literal["3.7"]
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS


class RangedStance_3_7(RangedStance_4_1):
    version: Literal["3.7"]


class MeleeStance_3_7(MeleeStance_4_1):
    version: Literal["3.7"]


class CeremonialGarment_3_5(RoundCharacterStatus):
    name: Literal["Ceremonial Garment"] = "Ceremonial Garment"
    version: Literal["3.5"] = "3.5"
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def value_modifier_DAMAGE_INCREASE(
        self,
        value: DamageIncreaseValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageIncreaseValue:
        """
        If self use normal attack, increase damage by 1.
        """
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not this character use normal attack, not modify
            return value
        # damage +1
        value.damage += 1
        return value

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        If self use normal attack, heal all our characters for 1 HP.
        """
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        ):
            # not self use skill
            return []
        skill = match.get_object(event.action.position)
        if skill.skill_type != SkillType.NORMAL_ATTACK:
            # not normal attack
            return []
        action = MakeDamageAction(damage_value_list=[])
        characters = match.player_tables[self.position.player_idx].characters
        for character in characters:
            if character.is_alive:
                action.damage_value_list.append(
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.HEAL,
                        target_position=character.position,
                        damage=-1,
                        damage_elemental_type=DamageElementalType.HEAL,
                        cost=Cost(),
                    )
                )
        return [action]


class HeronShield_3_8(ShieldCharacterStatus, PrepareCharacterStatus):
    name: Literal["Heron Shield"] = "Heron Shield"
    version: Literal["3.8"] = "3.8"
    character_name: Literal["Candace"] = "Candace"
    skill_name: Literal["Heron Strike"] = "Heron Strike"

    usage: int = 2
    max_usage: int = 2

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[Actions]:
        """
        Do not remove when usage becomes zero
        """
        return []


class Refraction_3_3(RoundCharacterStatus):
    name: Literal["Refraction"] = "Refraction"
    desc: Literal["", "talent"] = ""
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2

    is_talent_activated: bool = False
    cost_increase_usage: int = 1
    cost_increase_max_usage: int = 1
    icon_type: Literal[IconType.DEBUFF] = IconType.DEBUFF

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        if self.is_talent_activated:
            self.desc = "talent"

    def renew(self, new_status: "Refraction_3_3") -> None:
        if new_status.is_talent_activated:
            self.desc = "talent"
            self.is_talent_activated = True
        return super().renew(new_status)

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        self.cost_increase_usage = self.cost_increase_max_usage
        return super().event_handler_ROUND_PREPARE(event, match)

    def value_modifier_COST(
        self,
        value: CostValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> CostValue:
        """
        if talent activated, increase switch cost
        """
        if (
            self.is_talent_activated
            and self.cost_increase_usage > 0
            and value.cost.label & CostLabels.SWITCH_CHARACTER.value != 0
            and value.position.player_idx == self.position.player_idx
            and value.position.character_idx == self.position.character_idx
        ):
            # talent activated, switch cost, and this character switch to
            # other, increase cost
            value.cost.any_dice_number += 1
            if mode == "REAL":
                self.cost_increase_usage -= 1
        return value

    def value_modifier_DAMAGE_INCREASE(
        self,
        value: DamageIncreaseValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageIncreaseValue:
        if value.damage_elemental_type != DamageElementalType.HYDRO:
            # not hydro damage
            return value
        if value.is_corresponding_character_receive_damage(self.position, match):
            # this character receive damage
            value.damage += 1
        return value


class TakimeguriKanka_4_1(ElementalInfusionCharacterStatus, UsageCharacterStatus):
    name: Literal["Takimeguri Kanka"] = "Takimeguri Kanka"
    version: Literal["4.1"] = "4.1"
    usage: int = 2
    max_usage: int = 2
    infused_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    icon_type: Literal[IconType.ATK_UP_WATER] = IconType.ATK_UP_WATER

    def value_modifier_DAMAGE_INCREASE(
        self,
        value: DamageIncreaseValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageIncreaseValue:
        if value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            assert mode == "REAL"
            self.usage -= 1
            value.damage += 1
            character = match.player_tables[self.position.player_idx].characters[
                self.position.character_idx
            ]
            target = match.player_tables[value.target_position.player_idx].characters[
                value.target_position.character_idx
            ]
            if character.talent is not None and target.hp <= 6:
                value.damage += 1
        return value


class LingeringAeon_4_2(RoundEndAttackCharacterStatus):
    name: Literal["Lingering Aeon"] = "Lingering Aeon"
    version: Literal["4.2"] = "4.2"
    usage: int = 1
    max_usage: int = 1
    damage: int = 3
    damage_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS


register_class(
    Riptide_4_1
    | RangedStance_4_1
    | MeleeStance_4_1
    | Riptide_3_7
    | RangedStance_3_7
    | MeleeStance_3_7
    | CeremonialGarment_3_5
    | HeronShield_3_8
    | Refraction_3_3
    | TakimeguriKanka_4_1
    | LingeringAeon_4_2
)



from typing import Dict, List, Literal
from pydantic import PrivateAttr

from ....object_base import CreateSystemEventHandlerObject
from ....event_handler import SystemEventHandlerBase
from ....match import Match
from ....action import ChangeObjectUsageAction, CreateDiceAction, RemoveObjectAction
from ....event import (
    MoveObjectEventArguments,
    RemoveObjectEventArguments,
    SkillEndEventArguments,
)
from ....consts import DieColor, IconType, ObjectPositionType, SkillType
from ....card.support.companions import CompanionBase
from ....struct import Cost
from .....utils.class_registry import register_class
from .....utils.desc_registry import DescDictType


class JehtEventHandler_4_4(SystemEventHandlerBase):
    """
    It will be generated when Jeht in deck, and record disappeared support card numbers
    for both player, and when Jeht is played into area or any card is removed from
    support area, it will update its counter and also update all existing Jeht usage.
    """

    name: Literal["Jeht"] = "Jeht"
    version: Literal["4.4"] = "4.4"
    counters: List[int] = [0, 0]
    _max_usage: int = PrivateAttr(6)

    def _update_jedt_usage(self, match: Match) -> List[ChangeObjectUsageAction]:
        """
        check all jeht, if their usage not same as self, update it.
        """
        actions: List[ChangeObjectUsageAction] = []
        for counter, table in zip(self.counters, match.player_tables):
            for support in table.supports:
                if support.name == self.name and support.usage != counter:
                    actions.append(
                        ChangeObjectUsageAction(
                            object_position=support.position,
                            change_usage=counter - support.usage,
                        )
                    )
        return actions

    def event_handler_MOVE_OBJECT(
        self, event: MoveObjectEventArguments, match: Match
    ) -> List[ChangeObjectUsageAction]:
        if event.object_name != self.name:
            # not Jeht
            return []
        return self._update_jedt_usage(match)

    def event_handler_REMOVE_OBJECT(
        self, event: RemoveObjectEventArguments, match: Match
    ) -> List[ChangeObjectUsageAction]:
        if event.action.object_position.area != ObjectPositionType.SUPPORT:
            # not moved from support zone
            return []
        # add one usage
        pidx = event.action.object_position.player_idx
        self.counters[pidx] = min(self.counters[pidx] + 1, self._max_usage)
        return self._update_jedt_usage(match)


class Jeht_4_4(CreateSystemEventHandlerObject, CompanionBase):
    name: Literal["Jeht"]
    version: Literal["4.4"] = "4.4"
    usage: int = 0
    _max_usage: int = PrivateAttr(6)
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE
    cost: Cost = Cost(any_dice_number=2)

    handler_name: Literal["Jeht"] = "Jeht"

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Match
    ) -> List[CreateDiceAction | RemoveObjectAction]:
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        ):
            # not self use skill
            return []
        if self.usage < 5:
            # not enough usage
            return []
        if event.action.skill_type != SkillType.ELEMENTAL_BURST:
            # not use elemental burst
            return []
        # discard self and create dice
        return [
            CreateDiceAction(
                player_idx=self.position.player_idx,
                number=self.usage - 2,
                color=DieColor.OMNI,
            ),
            RemoveObjectAction(object_position=self.position),
        ]


desc: Dict[str, DescDictType] = {
    "SUPPORT/Jeht": {
        "names": {"en-US": "Jeht", "zh-CN": "婕德"},
        "descs": {
            "4.4": {
                "en-US": "This card will record the number of cards discarded from your Support Zone during this match as Sophistication points. (Max 6 points)\nAfter your characters use an Elemental Burst: If this card has recorded at least 5 Sophistication points, discard this card and generate Omni Element equal to the number of Sophistication points minus 2.",  # noqa: E501
                "zh-CN": "此牌会记录本场对局中我方支援区弃置卡牌的数量，称为「阅历」。（最多6点）\n我方角色使用「元素爆发」后：如果「阅历」至少为5，则弃置此牌，生成「阅历」-2数量的万能元素。",  # noqa: E501
            }
        },
        "image_path": "cardface/Assist_NPC_Jeht.png",  # noqa: E501
        "id": 322022,
    },
    "SYSTEM/Jeht": {
        "names": {"en-US": "Jeht", "zh-CN": "婕德"},
        "descs": {"4.4": {}},
    },
}


register_class(Jeht_4_4 | JehtEventHandler_4_4, desc)



from typing import Dict, List, Literal

from ....action import MakeDamageAction
from ....event import RoundEndEventArguments
from ....struct import Cost
from ....patch.v43.cards.vourukashas_glow_4_3 import VourukashasGlow_4_3
from ....match import Match
from .....utils import register_class, DescDictType


class VourukashasGlow_4_4(VourukashasGlow_4_3):
    name: Literal["Vourukasha's Glow"]
    version: Literal["4.4"] = "4.4"
    cost: Cost = Cost(same_dice_number=1)

    def event_handler_ROUND_END(
        self, desc: RoundEndEventArguments, match: Match
    ) -> List[MakeDamageAction]:
        if self.usage == self.max_usage_per_round:
            # not draw card, return
            return []
        return super().event_handler_ROUND_END(desc, match)


desc: Dict[str, DescDictType] = {
    "ARTIFACT/Vourukasha's Glow": {
        "descs": {
            "4.4": {
                "en-US": 'After this character takes DMG: Draw 1 card. (Once per Round)\nEnd Phase: If this character has triggered the effect "Draw 1 card" this round: heal the attached character for 1 HP.\n(A character can equip a maximum of 1 Artifact)',  # noqa: E501
                "zh-CN": "角色受到伤害后：如果所附属角色为「出战角色」，则抓1张牌。（每回合1次）\n结束阶段：若本角色在回合内触发了抓1张牌的效果，治疗所附属角色1点。\n（角色最多装备1件「圣遗物」）",  # noqa: E501
            }
        },
    },
}


register_class(VourukashasGlow_4_4, desc)



# TODO: cicin usage 2 get hutao should disappear
from typing import Any, Dict, List, Literal

from pydantic import PrivateAttr


from lpsim.server.character.cryo.fatui_cryo_cicin_mage_4_1 import CryoCicins_4_1
from lpsim.server.consts import IconType, PlayerActionLabels
from lpsim.server.status.character_status.base import PrepareCharacterStatus
from lpsim.server.status.team_status.base import ShieldTeamStatus, TeamStatusBase
from lpsim.utils.class_registry import register_class
from lpsim.server.match import Match
from lpsim.utils.desc_registry import DescDictType
from lpsim.server.modifiable_values import DamageValue
from lpsim.server.event import (
    ActionEndEventArguments,
    PlayerActionStartEventArguments,
    RemoveObjectEventArguments,
    RoundPrepareEventArguments,
    SkillEndEventArguments,
)
from lpsim.server.action import (
    ActionTypes,
    Actions,
    ChangeObjectUsageAction,
    CreateObjectAction,
    MakeDamageAction,
    RemoveObjectAction,
)
from lpsim.server.struct import Cost, ObjectPosition
from lpsim.consts import (
    ELEMENT_TO_DAMAGE_TYPE,
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    WeaponType,
)
from lpsim.server.character.character_base import (
    ElementalBurstBase,
    ElementalNormalAttackBase,
    ElementalSkillBase,
    CharacterBase,
    SkillTalent,
)


# Character status.


class ElectroCicinShield_4_5(ShieldTeamStatus):
    name: Literal["Electro Cicin Shield"] = "Electro Cicin Shield"
    version: Literal["4.5"] = "4.5"
    usage: int = 1
    max_usage: int = 1


# Team status.


class ElectroCicinStatus_4_5(TeamStatusBase):
    name: Literal["Electro Cicin"] = "Electro Cicin"
    version: Literal["4.5"] = "4.5"
    usage: int = 0
    max_usage: int = 3
    icon_type: IconType = IconType.DEBUFF

    def event_handler_ACTION_END(
        self, event: ActionEndEventArguments, match: Match
    ) -> List[ChangeObjectUsageAction]:
        """
        When play 3 cards, increase usage
        """
        if (
            event.action.position.player_idx != self.position.player_idx
            or event.action.action_label & PlayerActionLabels.CARD.value == 0
        ):
            # not self player, or not use card
            return []
        self.usage = min(self.max_usage, self.usage + 1)
        if self.usage != self.max_usage:
            return []
        cicin = self.query_one(match, 'opponent summon name="Electro Cicin"')
        assert cicin is not None
        if cicin.usage == 3:  # type: ignore
            # when cicin already full, not increase
            return []
        # increase cicin usage
        self.usage = 0
        return [ChangeObjectUsageAction(object_position=cicin.position, change_usage=1)]

    def event_handler_REMOVE_OBJECT(
        self, event: RemoveObjectEventArguments, match: Match
    ) -> List[RemoveObjectAction]:
        """
        when opposite cicin removed, remove self
        """
        if (
            self.position.not_satisfy(
                "both pidx=diff and target area=summon",
                target=event.action.object_position,
            )
            or event.object_name != "Electro Cicin"
        ):
            # not remove opposite electro cicin
            return []
        return [RemoveObjectAction(object_position=self.position)]


class SurgingThunderStatus_4_5(PrepareCharacterStatus):
    name: Literal["Surging Thunder"] = "Surging Thunder"
    version: Literal["4.5"] = "4.5"
    character_name: Literal["Fatui Electro Cicin Mage"] = "Fatui Electro Cicin Mage"
    skill_name: Literal["Surging Thunder"] = "Surging Thunder"


# Summons


class ElectroCicin_4_5(CryoCicins_4_1):
    name: Literal["Electro Cicin"] = "Electro Cicin"
    version: Literal["4.5"] = "4.5"
    usage: int = 3
    max_usage: int = 3
    damage_elemental_type: DamageElementalType = DamageElementalType.ELECTRO
    damage: int = 1
    _character_name: Literal["Fatui Electro Cicin Mage"] = PrivateAttr(
        "Fatui Electro Cicin Mage"
    )

    def renew(self, new_status: "ElectroCicin_4_5") -> None:
        # cryo cicin will over maximum, but electro will never.
        self.over_maximum = False
        return super().renew(new_status)

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[Actions]:
        # after skill end will do nothing
        return []


# Skills


class MistyCall(ElementalSkillBase):
    name: Literal["Misty Call"] = "Misty Call"
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=3)

    def get_actions(self, match: Match) -> List[Actions]:
        """
        Attack and create object
        """
        return [
            self.create_summon("Electro Cicin"),
            CreateObjectAction(
                object_name="Electro Cicin",
                object_position=ObjectPosition(
                    player_idx=1 - self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=-1,
                ),
                object_arguments={},
            ),
            self.charge_self(1),
        ]


class ThunderingShield(ElementalBurstBase):
    name: Literal["Thundering Shield"] = "Thundering Shield"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Match) -> List[Actions]:
        ret = super().get_actions(match)
        damage_action = ret[1]
        assert damage_action.type == ActionTypes.MAKE_DAMAGE
        sc = self.query_one(match, "self")
        assert sc is not None
        damage_action.damage_value_list.append(
            DamageValue.create_element_application(
                self.position,
                sc.position,
                element=DamageElementalType.ELECTRO,
                cost=self.cost,
            )
        )
        shield_args = {
            "usage": 1,
            "max_usage": 1,
        }
        cicin = self.query_one(match, 'our summon name="Electro Cicin"')
        if cicin is not None:
            usage = cicin.usage  # type: ignore
            shield_args["usage"] = 1 + usage
            shield_args["max_usage"] = 1 + usage
        ret.append(self.create_team_status("Electro Cicin Shield", shield_args))
        ret.append(self.create_character_status("Surging Thunder"))
        return ret


class SurgingThunder(ElementalBurstBase):
    name: Literal["Surging Thunder"] = "Surging Thunder"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost()

    def is_valid(self, match: Match) -> bool:
        """
        Always invalid for prepare skills
        """
        return False


# Talents


class ElectroCicinsGleam_4_5(SkillTalent):
    name: Literal["Electro Cicin's Gleam"]
    version: Literal["4.5"] = "4.5"
    character_name: Literal["Fatui Electro Cicin Mage"] = "Fatui Electro Cicin Mage"
    cost: Cost = Cost(elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=3)
    skill: Literal["Misty Call"] = "Misty Call"
    usage: int = 1

    def get_actions(self, target: ObjectPosition | None, match: Any) -> List[Actions]:
        self.usage = 1
        return super().get_actions(target, match)

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[Actions]:
        self.usage = 1
        return []

    def event_handler_PLAYER_ACTION_START(
        self, event: PlayerActionStartEventArguments, match: Match
    ) -> List[MakeDamageAction | ChangeObjectUsageAction]:
        # TODO: in game, when multiple talent exist, all of them trigger when condition
        # satisfied. Implement here will cause only one talent trigger, may move to
        # cicin.
        if self.usage == 0:
            return []
        if self.position.not_satisfy(f"source area=character pidx={event.player_idx}"):
            # not equipped, or not self start
            return []
        cicin: ElectroCicin_4_5 = self.query_one(
            match, 'our summon name="Electro Cicin"'
        )  # type: ignore
        if cicin is None or cicin.usage < 3:
            # no cicin or usage not full
            return []
        # attack and decrease usage
        target = self.query_one(match, "opponent active")
        assert target is not None
        self.usage -= 1
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=cicin.position,
                        damage_type=DamageType.DAMAGE,
                        target_position=target.position,
                        damage=cicin.damage,
                        damage_elemental_type=cicin.damage_elemental_type,
                        cost=Cost(),
                    )
                ]
            ),
            ChangeObjectUsageAction(object_position=cicin.position, change_usage=-1),
        ]


# character base


# character class name should contain its version.
class FatuiElectroCicinMage_4_5(CharacterBase):
    name: Literal["Fatui Electro Cicin Mage"]
    version: Literal["4.5"] = "4.5"
    element: ElementType = ElementType.ELECTRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        ElementalNormalAttackBase | MistyCall | ThunderingShield | SurgingThunder
    ] = []
    faction: List[FactionType] = [FactionType.MONSTER]
    weapon_type: WeaponType = WeaponType.OTHER

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name="Hurtling Bolts",
                damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost=ElementalNormalAttackBase.get_cost(self.element),
            ),
            MistyCall(),
            ThunderingShield(),
            SurgingThunder(),
        ]


# define descriptions of newly defined classes. Note key of skills and talents
# have character names. For balance changes, only descs are needed to define.
character_descs: Dict[str, DescDictType] = {
    "CHARACTER/Fatui Electro Cicin Mage": {
        "names": {"en-US": "Fatui Electro Cicin Mage", "zh-CN": "愚人众·雷萤术士"},
        "descs": {"4.5": {"en-US": "", "zh-CN": ""}},
        "image_path": "cardface/Char_Monster_FatuusSummoner.png",  # noqa: E501
        "id": 2404,
    },
    "SKILL_Fatui Electro Cicin Mage_NORMAL_ATTACK/Hurtling Bolts": {
        "names": {"en-US": "Hurtling Bolts", "zh-CN": "轰闪落雷"},
        "descs": {
            "4.5": {"en-US": "Deals 1 Electro DMG.", "zh-CN": "造成1点雷元素伤害。"}
        },
    },
    "SKILL_Fatui Electro Cicin Mage_ELEMENTAL_SKILL/Misty Call": {
        "names": {"en-US": "Misty Call", "zh-CN": "雾虚之召"},
        "descs": {"4.5": {"en-US": "Summons 1 Electro Cicin.", "zh-CN": "召唤雷萤。"}},
    },
    "TEAM_STATUS/Electro Cicin": {
        "names": {"en-US": "Electro Cicin", "zh-CN": "雷萤"},
        "descs": {
            "4.5": {
                "en-US": "After you play a total of 3 Action Cards: Your opponent's Electro Cicin gains 1 Usage(s).",  # noqa: E501
                "zh-CN": "我方累计打出3张行动牌后：敌方雷萤可用次数+1。",
            }
        },
    },
    "SUMMON/Electro Cicin": {
        "names": {"en-US": "Electro Cicin", "zh-CN": "雷萤"},
        "descs": {
            "4.5": {
                "en-US": "End Phase: Deal 1 Electro DMG.\nUsage(s): 3\n\nAfter your opponent plays a total of 3 Action Cards: This card gains 1 Usage(s). (Max 3 stacks)\nAfter Fatui Electro Cicin Mage takes Elemental Reaction DMG: This card loses 1 Usage(s).",  # noqa: E501
                "zh-CN": "结束阶段：造成1点雷元素伤害。\n可用次数：3\n\n敌方累计打出3张行动牌后：此牌可用次数+1。（最多叠加到3）\n愚人众·雷萤术士受到元素反应伤害后：此牌可用次数-1。",  # noqa: E501
            }
        },
        "image_path": "cardface/Summon_FatuusSummoner.png",
    },
    "SKILL_Fatui Electro Cicin Mage_ELEMENTAL_BURST/Thundering Shield": {
        "names": {"en-US": "Thundering Shield", "zh-CN": "霆雷之护"},
        "descs": {
            "4.5": {
                "en-US": "Deals 1 Electro DMG, applies Electro Application to this character, creates 1 Electro Cicin Shield and prepares Surging Thunder.",  # noqa: E501
                "zh-CN": "造成1点雷元素伤害，本角色附着雷元素，生成雷萤护罩并准备技能霆电迸发。",  # noqa: E501
            }
        },
    },
    "TEAM_STATUS/Electro Cicin Shield": {
        "names": {"en-US": "Electro Cicin Shield", "zh-CN": "雷萤护罩"},
        "descs": {
            "4.5": {
                "en-US": "Shield\nProvides 1 Shield point for your active character.\nWhen created: If you have Electro Cicin on the field, additionally increase Shield by the amount of Usage(s) it has. (Adds a maximum of 3 additional Shield)",  # noqa: E501
                "zh-CN": "护盾\n为我方出战角色提供1点护盾。\n创建时：如果我方场上存在雷萤，则额外提供其可用次数的护盾。（最多额外提供3点护盾）",  # noqa: E501
            }
        },
    },
    "CHARACTER_STATUS/Surging Thunder": {
        "names": {"en-US": "Surging Thunder", "zh-CN": "霆电迸发"},
        "descs": {
            "4.5": {
                "en-US": "Elemental Burst\n(Prepare for 1 turn)\nDeals 2 Electro DMG.",
                "zh-CN": "元素爆发\n（需准备1个行动轮）\n造成2点雷元素伤害。",
            }
        },
    },
    "SKILL_Fatui Electro Cicin Mage_ELEMENTAL_BURST/Surging Thunder": {
        "names": {"en-US": "Surging Thunder", "zh-CN": "霆电迸发"},
        "descs": {
            "4.5": {
                "en-US": "(Prepare for 1 turn)\nDeals 2 Electro DMG.",
                "zh-CN": "（需准备1个行动轮）\n造成2点雷元素伤害。",
            }
        },
    },
    "TALENT_Fatui Electro Cicin Mage/Electro Cicin's Gleam": {
        "names": {"en-US": "Electro Cicin's Gleam", "zh-CN": "雷萤浮闪"},
        "descs": {
            "4.5": {
                "en-US": "Combat Action: When your active character is Fatui Electro Cicin Mage, equip this card.\nAfter Fatui Electro Cicin Mage equips this card, immediately use Misty Call once.\nIf your Fatui Electro Cicin Mage who has this card equipped is on the field, before you choose an action: If Electro Cicin's Usage(s) is at least 3, then Electro Cicin will immediately deal 1 Electro DMG. (Usage(s) will be consumed, once per Round)\n(You must have Fatui Electro Cicin Mage in your deck to add this card to your deck.)",  # noqa: E501
                "zh-CN": "战斗行动：我方出战角色为愚人众·雷萤术士时，装备此牌。\n愚人众·雷萤术士装备此牌后，立刻使用一次雾虚之召。\n装备有此牌的愚人众·雷萤术士在场时，我方选择行动前：如果雷萤的可用次数至少为3，则雷萤立刻造成1点雷元素伤害。（需消耗可用次数，每回合1次）\n（牌组中包含愚人众·雷萤术士，才能加入牌组）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Talent_FatuusSummoner.png",  # noqa: E501
        "id": 224041,
    },
}


register_class(
    FatuiElectroCicinMage_4_5
    | ElectroCicin_4_5
    | ElectroCicinShield_4_5
    | ElectroCicinStatus_4_5
    | ElectroCicinsGleam_4_5
    | SurgingThunderStatus_4_5,
    character_descs,
)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...consts import (
    CostLabels,
    DamageElementalType,
    IconType,
    ObjectPositionType,
    SkillType,
)

from ...modifiable_values import CostValue, DamageIncreaseValue

from ...action import RemoveObjectAction, SkillEndAction

from ...event import MakeDamageEventArguments
from .base import (
    DefendCharacterStatus,
    RoundCharacterStatus,
    RoundEndAttackCharacterStatus,
    UsageCharacterStatus,
)


class Satiated_3_3(RoundCharacterStatus):
    name: Literal["Satiated"] = "Satiated"
    version: Literal["3.3"] = "3.3"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.FOOD] = IconType.FOOD


class JueyunGuoba_3_3(RoundCharacterStatus, UsageCharacterStatus):
    name: Literal["Jueyun Guoba"] = "Jueyun Guoba"
    version: Literal["3.3"] = "3.3"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        When this character use Elemental Burst, add 3 damage.

        Although it seems allow increase background damage, but as no
        skill will damage only background, it will be sonsumed when foreground
        character receives damage, and we do not need to check it.
        """
        assert mode == "REAL"
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not this character use elemental burst, not modify
            return value
        if self.usage <= 0:
            # no usage, not modify
            return value
        # elemental burst, modify
        value.damage += 1
        self.usage -= 1
        return value


class AdeptusTemptation_3_3(RoundCharacterStatus, UsageCharacterStatus):
    name: Literal["Adeptus' Temptation"] = "Adeptus' Temptation"
    version: Literal["3.3"] = "3.3"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        When this character use Elemental Burst, add 3 damage.

        Although it seems allow increase background damage, but as no
        skill will damage only background, it will be sonsumed when foreground
        character receives damage, and we do not need to check it.
        """
        assert mode == "REAL"
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.ELEMENTAL_BURST
        ):
            # not this character use elemental burst, not modify
            return value
        if self.usage <= 0:
            # no usage, not modify
            return value
        # elemental burst, modify
        value.damage += 3
        self.usage -= 1
        return value


class LotusFlowerCrisp_3_3(DefendCharacterStatus, RoundCharacterStatus):
    name: Literal["Lotus Flower Crisp"] = "Lotus Flower Crisp"
    version: Literal["3.3"] = "3.3"
    usage: int = 1
    max_usage: int = 1
    min_damage_to_trigger: int = 1
    max_in_one_time: int = 3


class NorthernSmokedChicken_3_3(RoundCharacterStatus, UsageCharacterStatus):
    name: Literal["Northern Smoked Chicken"]
    version: Literal["3.3"] = "3.3"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.BUFF] = IconType.BUFF

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        """
        If this character use normal attack, decrease one unaligned die cost.
        """
        if self.usage <= 0:  # pragma: no cover
            # no usage, not modify
            return value
        if not self.position.check_position_valid(
            value.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        ):
            # not character use skill, not modify
            return value
        if value.cost.label & CostLabels.NORMAL_ATTACK.value == 0:
            # not normal attack, not modify
            return value
        # modify
        if value.cost.decrease_cost(None):
            # decrease success
            if mode == "REAL":
                self.usage -= 1
        return value


class MushroomPizza_3_3(RoundEndAttackCharacterStatus):
    name: Literal["Mushroom Pizza"] = "Mushroom Pizza"
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.HEAL] = IconType.HEAL
    damage: int = -1
    damage_elemental_type: DamageElementalType = DamageElementalType.HEAL


class MintyMeatRolls_3_4(RoundCharacterStatus):
    name: Literal["Minty Meat Rolls"] = "Minty Meat Rolls"
    version: Literal["3.4"] = "3.4"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.BUFF] = IconType.BUFF

    decrease_usage: int = 3

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        """
        If this character use normal attack, decrease one unaligned die cost.
        """
        if self.decrease_usage <= 0:  # pragma: no cover
            # no usage, not modify
            return value
        if not self.position.check_position_valid(
            value.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        ):
            # not character use skill, not modify
            return value
        if value.cost.label & CostLabels.NORMAL_ATTACK.value == 0:
            # not normal attack, not modify
            return value
        # modify
        if value.cost.decrease_cost(None):
            # decrease success
            if mode == "REAL":
                self.decrease_usage -= 1
        return value

    def event_handler_SKILL_END(
        self, event: SkillEndAction, match: Any
    ) -> List[RemoveObjectAction]:
        if self.decrease_usage == 0:
            self.usage = 0
        return self.check_should_remove()


class MintyMeatRolls_3_3(MintyMeatRolls_3_4):
    version: Literal["3.3"]
    decrease_usage: int = 999


class SashiMiPlatter_3_7(RoundCharacterStatus):
    name: Literal["Sashimi Platter"] = "Sashimi Platter"
    version: Literal["3.7"] = "3.7"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        When this character use Elemental Burst, add 3 damage.

        Although it seems allow increase background damage, but as no
        skill will damage only background, it will be sonsumed when foreground
        character receives damage, and we do not need to check it.
        """
        assert mode == "REAL"
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not this character use elemental burst, not modify
            return value
        if value.damage_elemental_type == DamageElementalType.PIERCING:
            # piercing, not modify
            return value
        # elemental burst, modify
        value.damage += 1
        return value


class TandooriRoastChicken_3_7(RoundCharacterStatus):
    name: Literal["Tandoori Roast Chicken"] = "Tandoori Roast Chicken"
    version: Literal["3.7"] = "3.7"
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    usage: int = 1
    max_usage: int = 1

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        When this character use Elemental Skill, add 2 damage.
        Logic same as AdeptusTemptation.
        """
        assert mode == "REAL"
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.ELEMENTAL_SKILL
        ):
            # not this character use elemental skill, not modify
            return value
        if self.usage <= 0:  # pragma: no cover
            # no usage, not modify
            return value
        # elemental burst, modify
        value.damage += 2
        self.usage -= 1
        return value

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        return self.check_should_remove()


class ButterCrab_3_7(DefendCharacterStatus, RoundCharacterStatus):
    name: Literal["Butter Crab"] = "Butter Crab"
    version: Literal["3.7"] = "3.7"
    usage: int = 1
    max_usage: int = 1
    min_damage_to_trigger: int = 1
    max_in_one_time: int = 2


register_class(
    Satiated_3_3
    | JueyunGuoba_3_3
    | AdeptusTemptation_3_3
    | LotusFlowerCrisp_3_3
    | NorthernSmokedChicken_3_3
    | MushroomPizza_3_3
    | MintyMeatRolls_3_3
    | MintyMeatRolls_3_4
    | SashiMiPlatter_3_7
    | TandooriRoastChicken_3_7
    | ButterCrab_3_7
)
